#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commctrl.h>
#include <commdlg.h>
#include <richedit.h>
#include "ConvFunc.h"
#include "StackFunc.h"
#include "StrFunc.h"
#include "Resources\Resource.h"

#define AEC_FUNCTIONS
#include "AkelEdit.h"
#include "AkelDLL.h"


//Include conversion functions
#define xatoiA
#define xatoiW
#include "ConvFunc.h"

//Include stack functions
#define StackInsertAfter
#define StackInsertBefore
#define StackInsertIndex
#define StackDelete
#define StackClear
#define StackJoin
#include "StackFunc.h"

//Include string functions
#define WideCharUpper
#define xstrcmpiW
#define xstrcpynW
#define xstrstrW
#include "StrFunc.h"

//Defines
#define STRID_GOBEGIN                1
#define STRID_GOEND                  2
#define STRID_SELECT                 3
#define STRID_INVIEW                 4
#define STRID_LISTEXPAND             5
#define STRID_LISTCOLLAPSE           6
#define STRID_SETUP                  7
#define STRID_DOCKWINDOW             8
#define STRID_SHOWDOCK               9
#define STRID_FOLLOWCARET            10
#define STRID_OTHERSETTINGS          11
#define STRID_NOPRINTCOLLAPSED       12
#define STRID_COLORS                 13
#define STRID_FIRSTBACKGROUNDCOLOR   14
#define STRID_SECONDBACKGROUNDCOLOR  15
#define STRID_NORMALFOLDCOLOR        16
#define STRID_ACTIVEFOLDCOLOR        17
#define STRID_HOTKEYCURRENT          18
#define STRID_HOTKEYALL              19
#define STRID_COLLAPSE               20
#define STRID_NEXTLEVEL              21
#define STRID_PREVLEVEL              22
#define STRID_OK                     23
#define STRID_CANCEL                 24

#define BUFFER_SIZE      1024

//Dock messages
#define AKDLL_SETUP     (WM_USER + 1)
#define AKDLL_SETFOCUS  (WM_USER + 2)

//Option flags
#define OF_WINDOWRECT     0x1
#define OF_SETTINGS       0x2

#define DLLA_CODEFOLD_SETEXTENSION  1

//DestroyDock type
#define DKT_UNLOAD         0
#define DKT_NOUNLOAD       1
#define DKT_ONMAINFINISH   2

#define BOARD_WIDTH      10
#define BOARD_EDGE       4

//Painting figures
#define FRG_NONE            0
#define FRG_LINE            1
#define FRG_FOLDSTARTPLUS   2
#define FRG_FOLDSTARTMINUS  3
#define FRG_FOLDEND         4

//SetActiveEdit flags
#define SAE_RESETFOLD  0x1
#define SAE_RESETLIST  0x2

//IsFold return value
#define IF_FOLDSTART  0x1
#define IF_FOLDEND    0x2

//FOLDINFO flags
#define FIF_MATCHCASE             0x01
#define FIF_NAMEFROMLEFT          0x02
#define FIF_NAMEFROMRIGHT         0x04
#define FIF_ONLYFOLDSTART         0x08
#define FIF_FOLDSTARTATLINESTART  0x10
#define FIF_FOLDSTARTATLINEEND    0x20
#define FIF_FOLDENDATLINESTART    0x40
#define FIF_FOLDENDATLINEEND      0x80

typedef struct _EXTENSIONINFO {
  struct _EXTENSIONINFO *next;
  struct _EXTENSIONINFO *prev;
  wchar_t *wpExtension;
  int nExtensionLen;
} EXTENSIONINFO;

typedef struct _QUOTEINFO {
  struct _QUOTEINFO *next;
  struct _QUOTEINFO *prev;
  DWORD dwFlags;
  wchar_t *wpQuoteStart;
  int nQuoteStartLen;
  wchar_t *wpQuoteEnd;
  int nQuoteEndLen;
  wchar_t wchEscape;
} QUOTEINFO;

typedef struct _FOLDINFO {
  struct _FOLDINFO *next;
  struct _FOLDINFO *prev;
  DWORD dwFlags;
  wchar_t *wpFoldStart;
  int nFoldStartLen;
  wchar_t *wpFoldEnd;
  int nFoldEndLen;
  wchar_t *wpDelimiters;
} FOLDINFO;

typedef struct _HEXTENSIONSTACK {
  int first;
  int last;
  int lpExtensionLens[MAX_PATH];
} HEXTENSIONSTACK;

typedef struct _HQUOTESTACK {
  int first;
  int last;
} HQUOTESTACK;

typedef struct _HFOLDSTACK {
  int first;
  int last;
} HFOLDSTACK;

typedef struct _HSYNTAXFILE {
  struct _HSYNTAXFILE *next;
  struct _HSYNTAXFILE *prev;
  HEXTENSIONSTACK hExtensionStack;
  HQUOTESTACK hQuoteStack;
  HFOLDSTACK hFoldStack;
} HSYNTAXFILE;

typedef struct _LEVEL {
  struct _LEVEL *next;
  struct _LEVEL *prev;
  struct _LEVEL *lpParent;
  AEFOLD *lpFold;
  FOLDINFO *lpFoldInfo;
  HTREEITEM hItem;
  int nDeep;
  wchar_t wszName[MAX_PATH];
  AEPOINT pointMin;
  AEPOINT pointMax;
} LEVEL;

typedef struct _FOLDWINDOW {
  struct _FOLDWINDOW *next;
  struct _FOLDWINDOW *prev;
  HSYNTAXFILE *lpSyntaxFile;
  HSTACK hLevelStack;
  HSTACK hTempStack;
  HWND hWnd;
  RECT rcBoard;
  AECHARRANGE crTextChange;
  LEVEL *lpTextChangeMinParent;
  LEVEL *lpTextChangeMaxParent;
  BOOL bLevelChanged;
  BOOL bLockSel;
  BOOL bReturnPrintOption;
} FOLDWINDOW;

typedef struct _MANUALSET {
  struct _MANUALSET *next;
  struct _MANUALSET *prev;
  HWND hWnd;
  wchar_t wszExt[MAX_PATH];
} MANUALSET;

//LineBoard external call
typedef struct {
  DWORD dwStructSize;
  int nAction;
  HWND hWndEdit;
  RECT *rcBoard;
} DLLECLINEBOARD;

#define DLLA_LINEBOARD_CALCMAXWIDTH  1
#define DLLA_LINEBOARD_GETBOARDRECT  10

//DIALOGRESIZE
#define DRS_SIZE  0x1 //Resize control
#define DRS_MOVE  0x2 //Move control
#define DRS_X     0x4 //X value
#define DRS_Y     0x8 //Y value

typedef struct {
  HWND *lpWnd;   //Control window
  DWORD dwType;  //See DRS_* defines
} DIALOGRESIZE;

//Functions prototypes
void CreateDock(HWND *hWndDock, DOCK **dkDock);
void DestroyDock(HWND hWndDock, DWORD dwType);
BOOL CALLBACK MainDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK MainDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL DialogResizeMessages(DIALOGRESIZE *drs, RECT *rcInit, RECT *rcCurrent, HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL GetWindowPos(HWND hWnd, HWND hWndOwner, RECT *rc);
BOOL CALLBACK SetupDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK SetupDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK NewMainProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK NewFrameProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK EditParentMessages(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK NewEditProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK CurrentCollapseProc(void *lpParameter);
BOOL CALLBACK CurrentGoBeginProc(void *lpParameter);
BOOL CALLBACK CurrentGoEndProc(void *lpParameter);
BOOL CALLBACK CurrentSelectProc(void *lpParameter);
BOOL CALLBACK AllCollapseProc(void *lpParameter);
BOOL CALLBACK AllNextLevelProc(void *lpParameter);
BOOL CALLBACK AllPrevLevelProc(void *lpParameter);

HSYNTAXFILE* StackLoadSyntaxFile(HANDLE hFile, HSTACK *hStack, HSYNTAXFILE *lpSyntaxFile);
HSYNTAXFILE* StackInsertSyntaxFile(HSTACK *hStack);
HSYNTAXFILE* StackGetSyntaxFileByExt(HSTACK *hStack, wchar_t *wpExtension, int nExtensionLen);
void StackFreeSyntaxFile(HSTACK *hStack);
EXTENSIONINFO* StackInsertExtension(HEXTENSIONSTACK *hStack, int nExtensionLen);
EXTENSIONINFO* StackGetExtension(HEXTENSIONSTACK *hStack, wchar_t *wpExtension, int nExtensionLen);
void StackFreeExtension(HEXTENSIONSTACK *hStack);
MANUALSET* StackInsertManual(HSTACK *hStack);
MANUALSET* StackGetManual(HSTACK *hStack, HWND hWnd);
void StackDeleteManual(HSTACK *hStack, void *ms);
QUOTEINFO* StackInsertQuoteInfo(HQUOTESTACK *hStack);
void StackFreeQuoteInfo(HQUOTESTACK *hStack);
FOLDINFO* StackInsertFoldInfo(HFOLDSTACK *hStack);
void StackFreeFoldInfo(HFOLDSTACK *hStack);
void StackBeginBoards();
void StackEndBoards(HSTACK *hStack);
FOLDWINDOW* StackInsertFoldWindow(HSTACK *hStack);
FOLDWINDOW* StackGetFoldWindow(HSTACK *hStack, HWND hWnd);
void StackDeleteFoldWindow(HSTACK *hStack, FOLDWINDOW *lpElement);
void StackFreeFoldWindows(HSTACK *hStack);
LEVEL* StackInsertLevel(HSTACK *hLevelStack, AECHARINDEX *ciChar);
LEVEL* StackGetLevelByLine(HSTACK *hLevelStack, LEVEL **lpLevel, int nLine, BOOL bDeepest);
LEVEL* StackGetLevelByIndex(HSTACK *hLevelStack, LEVEL **lpLevel, AECHARINDEX *ciChar, BOOL bDeepest);
LEVEL* StackFoldAtIndex(HSTACK *hLevelStack, LEVEL *lpLevel, AECHARINDEX *ciChar, DWORD dwFoldStop);
BOOL StackIsLevelCollapsed(HSTACK *hLevelStack, LEVEL **lpLevel, int nLine);
void StackSwitchCollapse(HSTACK *hLevelStack, HWND hWnd, LEVEL *lpLevel);
void StackDeleteLevel(HSTACK *hLevelStack, HWND hWnd, LEVEL *lpLevel);
void StackFreeLevel(HSTACK *hLevelStack, HWND hWnd);
FOLDWINDOW* FillFoldsStack(FOLDWINDOW *lpFoldWindow, HSTACK *hLevelStack, HWND hWnd, AECHARRANGE *crRange);
void AddFoldsToStack(HSTACK *hLevelStack, HWND hWnd);
BOOL GetFoldName(const AECHARINDEX *ciInput, AECHARRANGE *crRange, wchar_t *wszName, int nMaxName);
BOOL FoldInView(FOLDWINDOW *lpFoldWindow, LEVEL *lpLevel, int nMenuAction);
BOOL FoldSelect(FOLDWINDOW *lpFoldWindow, LEVEL *lpLevel);
void FillTreeViewA(HWND hWndTreeView, FOLDWINDOW *lpFoldWindow);
void FillTreeViewW(HWND hWndTreeView, FOLDWINDOW *lpFoldWindow);
void FilterTreeViewA(HWND hWndTreeView, FOLDWINDOW *lpFoldWindow, char *pFindIt);
void FilterTreeViewW(HWND hWndTreeView, FOLDWINDOW *lpFoldWindow, wchar_t *wpFindIt);
void ExpandTreeView(HWND hWndTreeView, FOLDWINDOW *lpFoldWindow, BOOL bExpand);
LEVEL* GetCurLevelTreeView(HWND hWndTreeView);
BOOL SetCaretTreeView(HWND hWndTreeView, FOLDWINDOW *lpFoldWindow);
FOLDWINDOW* SetActiveEdit(HWND hWndEdit, HWND hWndTreeView, DWORD dwFlags);
DWORD IsMatch(AEFINDTEXTW *ft, const AECHARINDEX *ciChar);
BOOL IsEscaped(const AECHARINDEX *ciChar, wchar_t wchEscape);
FOLDINFO* IsFold(FOLDWINDOW *lpFoldWindow, LEVEL *lpLevel, AEFINDTEXTW *ft, AECHARINDEX *ciChar, DWORD *dwFoldStop);
FOLDINFO* FindFold(FOLDWINDOW *lpFoldWindow, LEVEL *lpStartScan, AECHARRANGE *crSearchRange);
BOOL CheckFlags(FOLDINFO *lpFoldInfo, AECHARRANGE *crFound, DWORD dwFoldStop);
BOOL AEC_IsSpacesFromLeft(const AECHARINDEX *ciChar);
BOOL AEC_IsSpacesFromRight(const AECHARINDEX *ciChar);
void SetEditRect(HWND hWnd, int nNewWidth, int nOldWidth);
HWND GetCurEdit();
DWORD ScrollToPoint(HWND hWnd, POINT *ptPos);
int GetWordA(char *pText, char *szWord, int nWordLenMax, char **pNextWord, BOOL *bQuote);
int GetWordW(wchar_t *wpText, wchar_t *wszWord, int nWordLenMax, wchar_t **wpNextWord, BOOL *bQuote);
BOOL NextLineA(char **pText);
BOOL NextLineW(wchar_t **wpText);
BOOL SkipCommentA(char **pText);
BOOL SkipCommentW(wchar_t **wpText);
BOOL IsDelimiter(const wchar_t *s, wchar_t c);
char* GetFileExtA(char *pFile);
wchar_t* GetFileExtW(wchar_t *wpFile);
void ReadSyntaxFilesA();
void ReadSyntaxFilesW();
void ReadOptionsA();
void ReadOptionsW();
void SaveOptionsA(DWORD dwFlags);
void SaveOptionsW(DWORD dwFlags);
char* GetLangStringA(LANGID wLangID, int nStringID);
wchar_t* GetLangStringW(LANGID wLangID, int nStringID);
BOOL IsExtCallParamValid(LPARAM lParam, int nIndex);
int GetExtCallParam(LPARAM lParam, int nIndex);
void InitCommon(PLUGINDATA *pd);
void InitMain();
void UninitMain();

//Global variables
char szPluginName[MAX_PATH];
wchar_t wszPluginName[MAX_PATH];
HINSTANCE hInstanceDLL;
HWND hMainWnd;
HWND hWndEdit;
HWND hMdiClient;
HICON hMainIcon;
BOOL bOldWindows;
BOOL bMDI;
LANGID wLangSystem;
BOOL bInitCommon=FALSE;
BOOL bInitMain=FALSE;
HWND hWndMainDlg=NULL;
RECT rcMainInitDialog={0};
RECT rcMainCurrentDialog={0};
RECT rcDockRect={0};
int nDockSide=DKS_RIGHT;
DOCK *dkMainDlg=NULL;
BOOL bDockWaitResize=FALSE;
BOOL bMainDialogRectSave=FALSE;
WNDPROCDATA *NewMainProcData=NULL;
WNDPROCDATA *NewFrameProcData=NULL;
WNDPROCDATA *NewEditProcData=NULL;
char szCodeFoldDir[MAX_PATH];
wchar_t wszCodeFoldDir[MAX_PATH];
char szBuffer[BUFFER_SIZE];
wchar_t wszBuffer[BUFFER_SIZE];
HSTACK hSyntaxFilesStack={0};
HSTACK hWindowsStack={0};
HSTACK hManualStack={0};
FOLDWINDOW *lpCurrentFoldWindow=NULL;
HWND hWndFunctionList=NULL;
BOOL bShowDock=TRUE;
BOOL bFollowCaret=TRUE;
BOOL bNoPrintCollapsed=FALSE;
COLORREF crPanelFirstBackground=RGB(0xFF, 0xFF, 0xFF);
COLORREF crPanelSecondBackground=RGB(0xE9, 0xE9, 0xE9);
COLORREF crPanelNormalFold=RGB(0x80, 0x80, 0x80);
COLORREF crPanelActiveFold=RGB(0x00, 0x00, 0xFF);
DWORD dwCurrentCollapse=699;  //"Ctrl+="
DWORD dwCurrentGoBegin=731;   //"Ctrl+["
DWORD dwCurrentGoEnd=733;     //"Ctrl+]"
DWORD dwCurrentSelect=732;    //"Ctrl+\"
DWORD dwAllCollapse=1723;     //"Ctrl+Alt+="
DWORD dwAllNextLevel=1757;    //"Ctrl+Alt+]"
DWORD dwAllPrevLevel=1755;    //"Ctrl+Alt+["
PLUGINFUNCTIONA *pfaCurrentCollapse=NULL;
PLUGINFUNCTIONA *pfaCurrentGoBegin=NULL;
PLUGINFUNCTIONA *pfaCurrentGoEnd=NULL;
PLUGINFUNCTIONA *pfaCurrentSelect=NULL;
PLUGINFUNCTIONA *pfaAllCollapse=NULL;
PLUGINFUNCTIONA *pfaAllNextLevel=NULL;
PLUGINFUNCTIONA *pfaAllPrevLevel=NULL;
PLUGINFUNCTIONW *pfwCurrentCollapse=NULL;
PLUGINFUNCTIONW *pfwCurrentGoBegin=NULL;
PLUGINFUNCTIONW *pfwCurrentGoEnd=NULL;
PLUGINFUNCTIONW *pfwCurrentSelect=NULL;
PLUGINFUNCTIONW *pfwAllCollapse=NULL;
PLUGINFUNCTIONW *pfwAllNextLevel=NULL;
PLUGINFUNCTIONW *pfwAllPrevLevel=NULL;

//Identification
extern "C" void __declspec(dllexport) DllAkelPadID(PLUGINVERSION *pv)
{
  pv->dwAkelDllVersion=AKELDLL;
  pv->dwExeMinVersion3x=MAKE_IDENTIFIER(-1, -1, -1, -1);
  pv->dwExeMinVersion4x=MAKE_IDENTIFIER(4, 4, 0, 0);
  pv->pPluginName="CodeFold";
}

//Plugin extern function
extern "C" void __declspec(dllexport) Main(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=TRUE;
    return;
  }

  if (!bInitCommon) InitCommon(pd);

  if (pd->lParam)
  {
    if (bInitMain)
    {
      if (IsExtCallParamValid(pd->lParam, 1))
      {
        if (GetExtCallParam(pd->lParam, 1) == DLLA_CODEFOLD_SETEXTENSION)
        {
          MANUALSET *ms;
          wchar_t *wpExt=NULL;

          if (IsExtCallParamValid(pd->lParam, 2))
            wpExt=(wchar_t *)GetExtCallParam(pd->lParam, 2);

          if (wpExt)
          {
            if (pd->hWndEdit)
            {
              if (!(ms=StackGetManual(&hManualStack, pd->hWndEdit)))
                ms=StackInsertManual(&hManualStack);

              if (ms)
              {
                ms->hWnd=pd->hWndEdit;
                xstrcpynW(ms->wszExt, wpExt, MAX_PATH);

                if (lpCurrentFoldWindow && lpCurrentFoldWindow->hWnd == pd->hWndEdit)
                  SetActiveEdit(pd->hWndEdit, hWndFunctionList, SAE_RESETFOLD|SAE_RESETLIST);
                else
                  SetActiveEdit(pd->hWndEdit, hWndFunctionList, SAE_RESETFOLD);
                InvalidateRect(pd->hWndEdit, NULL, TRUE);
              }
            }
          }
        }
      }

      //Stay in memory, and don't change active status
      pd->nUnload=UD_NONUNLOAD_UNCHANGE;
    }
  }
  else
  {
    if (bInitMain)
    {
      if (hWndMainDlg)
      {
        DestroyDock(hWndMainDlg, DKT_UNLOAD);

        //Stay in memory, but show as non-active
        pd->nUnload=UD_NONUNLOAD_NONACTIVE;
      }
      else
      {
        StackEndBoards(&hWindowsStack);
        StackFreeFoldWindows(&hWindowsStack);
        UninitMain();
      }
    }
    else
    {
      InitMain();

      if (bShowDock)
      {
        bDockWaitResize=pd->bOnStart;
        CreateDock(&hWndMainDlg, &dkMainDlg);
      }
      if (!pd->bOnStart)
      {
        if (hWndEdit=GetCurEdit())
        {
          StackBeginBoards();
          lpCurrentFoldWindow=SetActiveEdit(hWndEdit, hWndFunctionList, SAE_RESETLIST);
        }
      }

      //Stay in memory, and show as active
      pd->nUnload=UD_NONUNLOAD_ACTIVE;
    }
  }
}

extern "C" void __declspec(dllexport) Settings(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=FALSE;
    return;
  }

  if (!bInitCommon) InitCommon(pd);

  if (bOldWindows)
    DialogBoxA(hInstanceDLL, MAKEINTRESOURCEA(IDD_SETUP), hMainWnd, (DLGPROC)SetupDlgProcA);
  else
    DialogBoxW(hInstanceDLL, MAKEINTRESOURCEW(IDD_SETUP), hMainWnd, (DLGPROC)SetupDlgProcW);

  //If plugin already loaded, stay in memory, but show as non-active
  if (pd->bInMemory) pd->nUnload=UD_NONUNLOAD_NONACTIVE;
}

void CreateDock(HWND *hWndDock, DOCK **dkDock)
{
  DOCK dk={0};

  *hWndDock=NULL;
  dk.dwFlags=DKF_DRAGDROP;
  dk.nSide=nDockSide;
  dk.rcSize=rcDockRect;

  if (*dkDock=(DOCK *)SendMessage(hMainWnd, AKD_DOCK, DK_ADD, (LPARAM)&dk))
  {
    if (bOldWindows)
      *hWndDock=CreateDialogA(hInstanceDLL, MAKEINTRESOURCEA(IDD_LIST), hMainWnd, (DLGPROC)MainDlgProcA);
    else
      *hWndDock=CreateDialogW(hInstanceDLL, MAKEINTRESOURCEW(IDD_LIST), hMainWnd, (DLGPROC)MainDlgProcW);

    if (*hWndDock)
    {
      if (!(*dkDock)->rcSize.right || !(*dkDock)->rcSize.bottom)
      {
        GetWindowRect(*hWndDock, &(*dkDock)->rcSize);
        (*dkDock)->rcSize.right-=(*dkDock)->rcSize.left;
        (*dkDock)->rcSize.bottom-=(*dkDock)->rcSize.top;
        (*dkDock)->rcSize.left=0;
        (*dkDock)->rcSize.top=0;
      }
      (*dkDock)->hWnd=*hWndDock;
      SendMessage(hMainWnd, AKD_DOCK, DK_SUBCLASS, (LPARAM)*dkDock);

      if (!bDockWaitResize)
      {
        SendMessage(hMainWnd, AKD_RESIZE, 0, 0);
        ShowWindow(*hWndDock, SW_SHOW);
      }
    }
  }
}

void DestroyDock(HWND hWndDock, DWORD dwType)
{
  SendMessage(hWndDock, WM_COMMAND, IDC_CLOSE, dwType);
}

BOOL CALLBACK MainDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndTitleText;
  static HWND hWndTitleClose;
  static HWND hWndSearchText;
  static HMENU hMenuLabel;
  static HMENU hMenuTreeView;
  static DIALOGRESIZE drs[]={{&hWndTitleText,    DRS_SIZE|DRS_X},
                             {&hWndTitleClose,   DRS_MOVE|DRS_X},
                             {&hWndFunctionList, DRS_SIZE|DRS_X|DRS_Y},
                             {&hWndSearchText,   DRS_MOVE|DRS_Y},
                             {&hWndSearchText,   DRS_SIZE|DRS_X},
                             {0, 0}};

  if (uMsg == WM_INITDIALOG)
  {
    hWndFunctionList=GetDlgItem(hDlg, IDC_FUNCTIONLIST);
    hWndSearchText=GetDlgItem(hDlg, IDC_SEARCH);
    hWndTitleText=GetDlgItem(hDlg, IDC_TITLETEXT);
    hWndTitleClose=GetDlgItem(hDlg, IDC_CLOSE);

    hMenuLabel=CreatePopupMenu();
    AppendMenuA(hMenuLabel, MF_STRING, IDM_SETUP, GetLangStringA(wLangSystem, STRID_SETUP));

    hMenuTreeView=CreatePopupMenu();
    AppendMenuA(hMenuTreeView, MF_STRING, IDM_GOBEGIN, GetLangStringA(wLangSystem, STRID_GOBEGIN));
    AppendMenuA(hMenuTreeView, MF_STRING, IDM_GOEND, GetLangStringA(wLangSystem, STRID_GOEND));
    AppendMenuA(hMenuTreeView, MF_STRING, IDM_SELECT, GetLangStringA(wLangSystem, STRID_SELECT));
    AppendMenuA(hMenuTreeView, MF_STRING, IDM_INVIEW, GetLangStringA(wLangSystem, STRID_INVIEW));
    AppendMenuA(hMenuTreeView, MF_SEPARATOR, 0, NULL);
    AppendMenuA(hMenuTreeView, MF_STRING, IDM_LISTEXPAND, GetLangStringA(wLangSystem, STRID_LISTEXPAND));
    AppendMenuA(hMenuTreeView, MF_STRING, IDM_LISTCOLLAPSE, GetLangStringA(wLangSystem, STRID_LISTCOLLAPSE));
    AppendMenuA(hMenuTreeView, MF_SEPARATOR, 0, NULL);
    AppendMenuA(hMenuTreeView, MF_STRING, IDM_SETUP, GetLangStringA(wLangSystem, STRID_SETUP));
    SetMenuDefaultItem(hMenuTreeView, IDM_GOBEGIN, FALSE);

    SetWindowTextA(hWndTitleText, szPluginName);
    SendMessage(hMainWnd, AKD_SETCLOSEBUTTON, (WPARAM)hWndTitleClose, (LPARAM)NULL);
  }
  else if (uMsg == AKDLL_SETUP)
  {
    DialogBoxA(hInstanceDLL, MAKEINTRESOURCEA(IDD_SETUP), hDlg, (DLGPROC)SetupDlgProcA);
  }
  else if (uMsg == AKDLL_SETFOCUS)
  {
    if (lpCurrentFoldWindow)
    {
      SetFocus(lpCurrentFoldWindow->hWnd);
    }
  }
  else if (uMsg == WM_LBUTTONDBLCLK)
  {
    POINT pt;
    RECT rc;

    GetCursorPos(&pt);
    GetWindowRect(hWndTitleText, &rc);

    if (PtInRect(&rc, pt))
    {
      PostMessage(hDlg, AKDLL_SETUP, 0, 0);
    }
  }
  else if (uMsg == WM_CONTEXTMENU)
  {
    if ((HWND)wParam == hDlg)
    {
      RECT rc;
      POINT pt;

      GetCursorPos(&pt);
      GetWindowRect(hWndTitleText, &rc);

      if (PtInRect(&rc, pt))
      {
        int nCmd;

        nCmd=TrackPopupMenu(hMenuLabel, TPM_RETURNCMD|TPM_NONOTIFY|TPM_LEFTBUTTON|TPM_RIGHTBUTTON, pt.x, pt.y, 0, hWndTitleText, NULL);

        if (nCmd == IDM_SETUP)
          PostMessage(hDlg, AKDLL_SETUP, 0, 0);
      }
    }
    else if ((HWND)wParam == hWndFunctionList)
    {
      TVHITTESTINFO tvhti;
      RECT rcItem;
      POINT pt;
      int nCmd;

      if (lpCurrentFoldWindow)
      {
        if (lParam == -1)
        {
          *(HTREEITEM *)&rcItem=(HTREEITEM)SendMessage(hWndFunctionList, TVM_GETNEXTITEM, TVGN_CARET, (LPARAM)NULL);
          SendMessage(hWndFunctionList, TVM_GETITEMRECT, TRUE, (LPARAM)&rcItem);
          pt.x=rcItem.left;
          pt.y=rcItem.bottom;
          ClientToScreen(hWndFunctionList, &pt);
        }
        else
        {
          GetCursorPos(&tvhti.pt);
          ScreenToClient(hWndFunctionList, &tvhti.pt);
          SendMessage(hWndFunctionList, TVM_HITTEST, 0, (LPARAM)&tvhti);
          if (tvhti.hItem && (tvhti.flags & TVHT_ONITEM))
            SendMessage(hWndFunctionList, TVM_SELECTITEM, TVGN_CARET, (LPARAM)tvhti.hItem);
          GetCursorPos(&pt);
        }

        nCmd=TrackPopupMenu(hMenuTreeView, TPM_RETURNCMD|TPM_NONOTIFY|TPM_LEFTBUTTON|TPM_RIGHTBUTTON, pt.x, pt.y, 0, hWndFunctionList, NULL);

        if (nCmd == IDM_GOBEGIN ||
            nCmd == IDM_GOEND ||
            nCmd == IDM_INVIEW)
        {
          LEVEL *lpLevel;

          if (lpLevel=GetCurLevelTreeView(hWndFunctionList))
          {
            if (FoldInView(lpCurrentFoldWindow, lpLevel, nCmd))
              PostMessage(hWndMainDlg, AKDLL_SETFOCUS, 0, 0);
          }
        }
        else if (nCmd == IDM_SELECT)
        {
          LEVEL *lpLevel;

          if (lpLevel=GetCurLevelTreeView(hWndFunctionList))
          {
            if (FoldSelect(lpCurrentFoldWindow, lpLevel))
              PostMessage(hWndMainDlg, AKDLL_SETFOCUS, 0, 0);
          }
        }
        else if (nCmd == IDM_LISTEXPAND)
        {
          ExpandTreeView(hWndFunctionList, lpCurrentFoldWindow, TRUE);
        }
        else if (nCmd == IDM_LISTCOLLAPSE)
        {
          ExpandTreeView(hWndFunctionList, lpCurrentFoldWindow, FALSE);
        }
        else if (nCmd == IDM_SETUP)
        {
          PostMessage(hDlg, AKDLL_SETUP, 0, 0);
        }
      }
    }
  }
  else if (uMsg == WM_NOTIFY)
  {
    if ((int)((NMHDR *)lParam)->code == NM_DBLCLK)
    {
      TVHITTESTINFO tvhti;

      if (lpCurrentFoldWindow)
      {
        GetCursorPos(&tvhti.pt);
        ScreenToClient(hWndFunctionList, &tvhti.pt);
        SendMessage(hWndFunctionList, TVM_HITTEST, 0, (LPARAM)&tvhti);
        if (tvhti.hItem && (tvhti.flags & TVHT_ONITEM))
        {
          LEVEL *lpLevel;

          if (lpLevel=GetCurLevelTreeView(hWndFunctionList))
          {
            if (FoldInView(lpCurrentFoldWindow, lpLevel, IDM_GOBEGIN))
              PostMessage(hWndMainDlg, AKDLL_SETFOCUS, 0, 0);
          }
          SetWindowLongA(hDlg, DWL_MSGRESULT, 1);
          return 1;
        }
      }
    }
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_SEARCH)
    {
      if (HIWORD(wParam) == EN_CHANGE)
      {
        if (lpCurrentFoldWindow)
        {
          char szSearch[MAX_PATH];

          if (GetWindowTextA(hWndSearchText, szSearch, MAX_PATH))
            FilterTreeViewA(hWndFunctionList, lpCurrentFoldWindow, szSearch);
          else
            FillTreeViewA(hWndFunctionList, lpCurrentFoldWindow);
        }
      }
    }
    else if (LOWORD(wParam) == IDC_CLOSE ||
             LOWORD(wParam) == IDCANCEL)
    {
      if (bMainDialogRectSave || dkMainDlg->nSide != nDockSide)
        SaveOptionsA(OF_WINDOWRECT);

      if (lParam != DKT_NOUNLOAD)
      {
        if (lParam != DKT_ONMAINFINISH)
          StackEndBoards(&hWindowsStack);
        StackFreeFoldWindows(&hWindowsStack);
        UninitMain();
      }
      SendMessage(hMainWnd, AKD_DOCK, DK_DELETE, (LPARAM)dkMainDlg);
      dkMainDlg=NULL;

      DestroyMenu(hMenuLabel);
      DestroyMenu(hMenuTreeView);
      DestroyWindow(hWndMainDlg);
      hWndMainDlg=NULL;

      if (lParam != DKT_ONMAINFINISH)
      {
        SendMessage(hMainWnd, AKD_RESIZE, 0, 0);
        SetFocus(hMainWnd);
        if (lParam != DKT_NOUNLOAD)
          PostMessage(hMainWnd, AKD_DLLUNLOAD, (WPARAM)hInstanceDLL, (LPARAM)NULL);
      }
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }
  if (DialogResizeMessages(&drs[0], &rcMainInitDialog, &rcMainCurrentDialog, hDlg, uMsg, wParam, lParam))
  {
    if (IsWindowVisible(hDlg)) bMainDialogRectSave=TRUE;
    if (dkMainDlg) GetWindowPos(hWndTitleText, hDlg, &dkMainDlg->rcDragDrop);
  }

  return FALSE;
}

BOOL CALLBACK MainDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndTitleText;
  static HWND hWndTitleClose;
  static HWND hWndSearchText;
  static HMENU hMenuLabel;
  static HMENU hMenuTreeView;
  static DIALOGRESIZE drs[]={{&hWndTitleText,    DRS_SIZE|DRS_X},
                             {&hWndTitleClose,   DRS_MOVE|DRS_X},
                             {&hWndFunctionList, DRS_SIZE|DRS_X|DRS_Y},
                             {&hWndSearchText,   DRS_MOVE|DRS_Y},
                             {&hWndSearchText,   DRS_SIZE|DRS_X},
                             {0, 0}};

  if (uMsg == WM_INITDIALOG)
  {
    hWndFunctionList=GetDlgItem(hDlg, IDC_FUNCTIONLIST);
    hWndSearchText=GetDlgItem(hDlg, IDC_SEARCH);
    hWndTitleText=GetDlgItem(hDlg, IDC_TITLETEXT);
    hWndTitleClose=GetDlgItem(hDlg, IDC_CLOSE);

    hMenuLabel=CreatePopupMenu();
    AppendMenuW(hMenuLabel, MF_STRING, IDM_SETUP, GetLangStringW(wLangSystem, STRID_SETUP));

    hMenuTreeView=CreatePopupMenu();
    AppendMenuW(hMenuTreeView, MF_STRING, IDM_GOBEGIN, GetLangStringW(wLangSystem, STRID_GOBEGIN));
    AppendMenuW(hMenuTreeView, MF_STRING, IDM_GOEND, GetLangStringW(wLangSystem, STRID_GOEND));
    AppendMenuW(hMenuTreeView, MF_STRING, IDM_SELECT, GetLangStringW(wLangSystem, STRID_SELECT));
    AppendMenuW(hMenuTreeView, MF_STRING, IDM_INVIEW, GetLangStringW(wLangSystem, STRID_INVIEW));
    AppendMenuW(hMenuTreeView, MF_SEPARATOR, 0, NULL);
    AppendMenuW(hMenuTreeView, MF_STRING, IDM_LISTEXPAND, GetLangStringW(wLangSystem, STRID_LISTEXPAND));
    AppendMenuW(hMenuTreeView, MF_STRING, IDM_LISTCOLLAPSE, GetLangStringW(wLangSystem, STRID_LISTCOLLAPSE));
    AppendMenuW(hMenuTreeView, MF_SEPARATOR, 0, NULL);
    AppendMenuW(hMenuTreeView, MF_STRING, IDM_SETUP, GetLangStringW(wLangSystem, STRID_SETUP));
    SetMenuDefaultItem(hMenuTreeView, IDM_GOBEGIN, FALSE);

    SetWindowTextW(hWndTitleText, wszPluginName);
    SendMessage(hMainWnd, AKD_SETCLOSEBUTTON, (WPARAM)hWndTitleClose, (LPARAM)NULL);
  }
  else if (uMsg == AKDLL_SETUP)
  {
    DialogBoxW(hInstanceDLL, MAKEINTRESOURCEW(IDD_SETUP), hDlg, (DLGPROC)SetupDlgProcW);
  }
  else if (uMsg == AKDLL_SETFOCUS)
  {
    if (lpCurrentFoldWindow)
    {
      SetFocus(lpCurrentFoldWindow->hWnd);
    }
  }
  else if (uMsg == WM_LBUTTONDBLCLK)
  {
    POINT pt;
    RECT rc;

    GetCursorPos(&pt);
    GetWindowRect(hWndTitleText, &rc);

    if (PtInRect(&rc, pt))
    {
      PostMessage(hDlg, AKDLL_SETUP, 0, 0);
    }
  }
  else if (uMsg == WM_CONTEXTMENU)
  {
    if ((HWND)wParam == hDlg)
    {
      RECT rc;
      POINT pt;

      GetCursorPos(&pt);
      GetWindowRect(hWndTitleText, &rc);

      if (PtInRect(&rc, pt))
      {
        int nCmd;

        nCmd=TrackPopupMenu(hMenuLabel, TPM_RETURNCMD|TPM_NONOTIFY|TPM_LEFTBUTTON|TPM_RIGHTBUTTON, pt.x, pt.y, 0, hWndTitleText, NULL);

        if (nCmd == IDM_SETUP)
          PostMessage(hDlg, AKDLL_SETUP, 0, 0);
      }
    }
    else if ((HWND)wParam == hWndFunctionList)
    {
      TVHITTESTINFO tvhti;
      RECT rcItem;
      POINT pt;
      int nCmd;

      if (lpCurrentFoldWindow)
      {
        if (lParam == -1)
        {
          *(HTREEITEM *)&rcItem=(HTREEITEM)SendMessage(hWndFunctionList, TVM_GETNEXTITEM, TVGN_CARET, (LPARAM)NULL);
          SendMessage(hWndFunctionList, TVM_GETITEMRECT, TRUE, (LPARAM)&rcItem);
          pt.x=rcItem.left;
          pt.y=rcItem.bottom;
          ClientToScreen(hWndFunctionList, &pt);
        }
        else
        {
          GetCursorPos(&tvhti.pt);
          ScreenToClient(hWndFunctionList, &tvhti.pt);
          SendMessage(hWndFunctionList, TVM_HITTEST, 0, (LPARAM)&tvhti);
          if (tvhti.hItem && (tvhti.flags & TVHT_ONITEM))
            SendMessage(hWndFunctionList, TVM_SELECTITEM, TVGN_CARET, (LPARAM)tvhti.hItem);
          GetCursorPos(&pt);
        }

        nCmd=TrackPopupMenu(hMenuTreeView, TPM_RETURNCMD|TPM_NONOTIFY|TPM_LEFTBUTTON|TPM_RIGHTBUTTON, pt.x, pt.y, 0, hWndFunctionList, NULL);

        if (nCmd == IDM_GOBEGIN ||
            nCmd == IDM_GOEND ||
            nCmd == IDM_INVIEW)
        {
          LEVEL *lpLevel;

          if (lpLevel=GetCurLevelTreeView(hWndFunctionList))
          {
            if (FoldInView(lpCurrentFoldWindow, lpLevel, nCmd))
              PostMessage(hWndMainDlg, AKDLL_SETFOCUS, 0, 0);
          }
        }
        else if (nCmd == IDM_SELECT)
        {
          LEVEL *lpLevel;

          if (lpLevel=GetCurLevelTreeView(hWndFunctionList))
          {
            if (FoldSelect(lpCurrentFoldWindow, lpLevel))
              PostMessage(hWndMainDlg, AKDLL_SETFOCUS, 0, 0);
          }
        }
        else if (nCmd == IDM_LISTEXPAND)
        {
          ExpandTreeView(hWndFunctionList, lpCurrentFoldWindow, TRUE);
        }
        else if (nCmd == IDM_LISTCOLLAPSE)
        {
          ExpandTreeView(hWndFunctionList, lpCurrentFoldWindow, FALSE);
        }
        else if (nCmd == IDM_SETUP)
        {
          PostMessage(hDlg, AKDLL_SETUP, 0, 0);
        }
      }
    }
  }
  else if (uMsg == WM_NOTIFY)
  {
    if ((int)((NMHDR *)lParam)->code == NM_DBLCLK)
    {
      TVHITTESTINFO tvhti;

      if (lpCurrentFoldWindow)
      {
        GetCursorPos(&tvhti.pt);
        ScreenToClient(hWndFunctionList, &tvhti.pt);
        SendMessage(hWndFunctionList, TVM_HITTEST, 0, (LPARAM)&tvhti);
        if (tvhti.hItem && (tvhti.flags & TVHT_ONITEM))
        {
          LEVEL *lpLevel;

          if (lpLevel=GetCurLevelTreeView(hWndFunctionList))
          {
            if (FoldInView(lpCurrentFoldWindow, lpLevel, IDM_GOBEGIN))
              PostMessage(hWndMainDlg, AKDLL_SETFOCUS, 0, 0);
          }
          SetWindowLongA(hDlg, DWL_MSGRESULT, 1);
          return 1;
        }
      }
    }
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_SEARCH)
    {
      if (HIWORD(wParam) == EN_CHANGE)
      {
        if (lpCurrentFoldWindow)
        {
          wchar_t wszSearch[MAX_PATH];

          if (GetWindowTextW(hWndSearchText, wszSearch, MAX_PATH))
            FilterTreeViewW(hWndFunctionList, lpCurrentFoldWindow, wszSearch);
          else
            FillTreeViewW(hWndFunctionList, lpCurrentFoldWindow);
        }
      }
    }
    else if (LOWORD(wParam) == IDC_CLOSE ||
             LOWORD(wParam) == IDCANCEL)
    {
      if (bMainDialogRectSave || dkMainDlg->nSide != nDockSide)
        SaveOptionsW(OF_WINDOWRECT);

      if (lParam != DKT_NOUNLOAD)
      {
        if (lParam != DKT_ONMAINFINISH)
          StackEndBoards(&hWindowsStack);
        StackFreeFoldWindows(&hWindowsStack);
        UninitMain();
      }
      SendMessage(hMainWnd, AKD_DOCK, DK_DELETE, (LPARAM)dkMainDlg);
      dkMainDlg=NULL;

      DestroyMenu(hMenuLabel);
      DestroyMenu(hMenuTreeView);
      DestroyWindow(hWndMainDlg);
      hWndMainDlg=NULL;

      if (lParam != DKT_ONMAINFINISH)
      {
        SendMessage(hMainWnd, AKD_RESIZE, 0, 0);
        SetFocus(hMainWnd);
        if (lParam != DKT_NOUNLOAD)
          PostMessage(hMainWnd, AKD_DLLUNLOAD, (WPARAM)hInstanceDLL, (LPARAM)NULL);
      }
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }
  if (DialogResizeMessages(&drs[0], &rcMainInitDialog, &rcMainCurrentDialog, hDlg, uMsg, wParam, lParam))
  {
    if (IsWindowVisible(hDlg)) bMainDialogRectSave=TRUE;
    if (dkMainDlg) GetWindowPos(hWndTitleText, hDlg, &dkMainDlg->rcDragDrop);
  }

  return FALSE;
}

BOOL DialogResizeMessages(DIALOGRESIZE *drs, RECT *rcInit, RECT *rcCurrent, HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == WM_INITDIALOG)
  {
    RECT rcTemplate;

    GetWindowPos(hDlg, NULL, rcInit);
    rcTemplate=*rcCurrent;
    *rcCurrent=*rcInit;

    if (rcTemplate.right && rcTemplate.bottom)
    {
      rcTemplate.left=rcInit->left + (rcInit->right - rcTemplate.right) / 2;
      rcTemplate.top=rcInit->top + (rcInit->bottom - rcTemplate.bottom) / 2;
      SetWindowPos(hDlg, 0, rcTemplate.left, rcTemplate.top, rcTemplate.right, rcTemplate.bottom, SWP_NOZORDER);
    }
  }
  else if (uMsg == WM_GETMINMAXINFO)
  {
    //MINMAXINFO *mmi=(MINMAXINFO *)lParam;
    //
    //mmi->ptMinTrackSize.x=rcInit->right;
    //mmi->ptMinTrackSize.y=rcInit->bottom;
  }
  else if (uMsg == WM_SIZE)
  {
    if (lParam)
    {
      RECT rcTemplate;
      RECT rcControl;
      DWORD dwFlags;
      POINT pt;
      int i;

      GetWindowPos(hDlg, NULL, &rcTemplate);
      pt.x=rcTemplate.right - rcCurrent->right;
      pt.y=rcTemplate.bottom - rcCurrent->bottom;
      *rcCurrent=rcTemplate;

      for (i=0; drs[i].lpWnd; ++i)
      {
        if (*drs[i].lpWnd)
        {
          dwFlags=0;
          if (drs[i].dwType & DRS_SIZE)
            dwFlags|=SWP_NOMOVE;
          else if (drs[i].dwType & DRS_MOVE)
            dwFlags|=SWP_NOSIZE;
          GetWindowPos(*drs[i].lpWnd, hDlg, &rcControl);
          SetWindowPos(*drs[i].lpWnd, 0, rcControl.left + ((drs[i].dwType & DRS_X)?pt.x:0), rcControl.top + ((drs[i].dwType & DRS_Y)?pt.y:0), rcControl.right + ((drs[i].dwType & DRS_X)?pt.x:0), rcControl.bottom + ((drs[i].dwType & DRS_Y)?pt.y:0), dwFlags|SWP_NOZORDER);
        }
      }
      InvalidateRect(hDlg, NULL, TRUE);
      return TRUE;
    }
  }
  else if (uMsg == WM_PAINT)
  {
    //PAINTSTRUCT ps;
    //RECT rcGrip;
    //HDC hDC;
    //
    //if (hDC=BeginPaint(hDlg, &ps))
    //{
    //  GetClientRect(hDlg, &rcGrip);
    //  rcGrip.left=rcGrip.right - GetSystemMetrics(SM_CXVSCROLL);
    //  rcGrip.top=rcGrip.bottom - GetSystemMetrics(SM_CYVSCROLL);
    //  DrawFrameControl(hDC, &rcGrip, DFC_SCROLL, DFCS_SCROLLSIZEGRIP);
    //  EndPaint(hDlg, &ps);
    //}
  }
  return FALSE;
}

BOOL GetWindowPos(HWND hWnd, HWND hWndOwner, RECT *rc)
{
  if (GetWindowRect(hWnd, rc))
  {
    rc->right-=rc->left;
    rc->bottom-=rc->top;

    if (hWndOwner)
    {
      if (!ScreenToClient(hWndOwner, (POINT *)&rc->left))
        return FALSE;
    }
    return TRUE;
  }
  return FALSE;
}

BOOL CALLBACK SetupDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndShowDock;
  static HWND hWndFollowCaret;
  static HWND hWndNoPrintCollapsed;
  static HWND hWndFirstBackgroundColor;
  static HWND hWndSecondBackgroundColor;
  static HWND hWndNormalFoldColor;
  static HWND hWndActiveFoldColor;
  static HWND hWndCurrentCollapse;
  static HWND hWndCurrentGoBegin;
  static HWND hWndCurrentGoEnd;
  static HWND hWndCurrentSelect;
  static HWND hWndAllCollapse;
  static HWND hWndAllNextLevel;
  static HWND hWndAllPrevLevel;
  static COLORREF crPanelFirstBackgroundDlg;
  static COLORREF crPanelSecondBackgroundDlg;
  static COLORREF crPanelNormalFoldDlg;
  static COLORREF crPanelActiveFoldDlg;

  if (uMsg == WM_INITDIALOG)
  {
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hMainIcon);
    hWndShowDock=GetDlgItem(hDlg, IDC_SETUP_SHOWDOCK);
    hWndFollowCaret=GetDlgItem(hDlg, IDC_SETUP_FOLLOWCARET);
    hWndNoPrintCollapsed=GetDlgItem(hDlg, IDC_SETUP_NOPRINTCOLLAPSED);
    hWndFirstBackgroundColor=GetDlgItem(hDlg, IDC_SETUP_FIRSTBACKGROUNDCOLOR_BUTTON);
    hWndSecondBackgroundColor=GetDlgItem(hDlg, IDC_SETUP_SECONDBACKGROUNDCOLOR_BUTTON);
    hWndNormalFoldColor=GetDlgItem(hDlg, IDC_SETUP_NORMALFOLDCOLOR_BUTTON);
    hWndActiveFoldColor=GetDlgItem(hDlg, IDC_SETUP_ACTIVEFOLDCOLOR_BUTTON);
    hWndCurrentCollapse=GetDlgItem(hDlg, IDC_SETUP_HOTKEYCURRENT_COLLAPSE);
    hWndCurrentGoBegin=GetDlgItem(hDlg, IDC_SETUP_HOTKEYCURRENT_GOBEGIN);
    hWndCurrentGoEnd=GetDlgItem(hDlg, IDC_SETUP_HOTKEYCURRENT_GOEND);
    hWndCurrentSelect=GetDlgItem(hDlg, IDC_SETUP_HOTKEYCURRENT_SELECT);
    hWndAllCollapse=GetDlgItem(hDlg, IDC_SETUP_HOTKEYALL_COLLAPSE);
    hWndAllNextLevel=GetDlgItem(hDlg, IDC_SETUP_HOTKEYALL_NEXTLEVEL);
    hWndAllPrevLevel=GetDlgItem(hDlg, IDC_SETUP_HOTKEYALL_PREVLEVEL);

    SetDlgItemTextA(hDlg, IDC_SETUP_DOCK_GROUP, GetLangStringA(wLangSystem, STRID_DOCKWINDOW));
    SetDlgItemTextA(hDlg, IDC_SETUP_SHOWDOCK, GetLangStringA(wLangSystem, STRID_SHOWDOCK));
    SetDlgItemTextA(hDlg, IDC_SETUP_FOLLOWCARET, GetLangStringA(wLangSystem, STRID_FOLLOWCARET));
    SetDlgItemTextA(hDlg, IDC_SETUP_OTHER_GROUP, GetLangStringA(wLangSystem, STRID_OTHERSETTINGS));
    SetDlgItemTextA(hDlg, IDC_SETUP_NOPRINTCOLLAPSED, GetLangStringA(wLangSystem, STRID_NOPRINTCOLLAPSED));
    SetDlgItemTextA(hDlg, IDC_SETUP_COLORS_GROUP, GetLangStringA(wLangSystem, STRID_COLORS));
    SetDlgItemTextA(hDlg, IDC_SETUP_FIRSTBACKGROUNDCOLOR_LABEL, GetLangStringA(wLangSystem, STRID_FIRSTBACKGROUNDCOLOR));
    SetDlgItemTextA(hDlg, IDC_SETUP_SECONDBACKGROUNDCOLOR_LABEL, GetLangStringA(wLangSystem, STRID_SECONDBACKGROUNDCOLOR));
    SetDlgItemTextA(hDlg, IDC_SETUP_NORMALFOLDCOLOR_LABEL, GetLangStringA(wLangSystem, STRID_NORMALFOLDCOLOR));
    SetDlgItemTextA(hDlg, IDC_SETUP_ACTIVEFOLDCOLOR_LABEL, GetLangStringA(wLangSystem, STRID_ACTIVEFOLDCOLOR));
    SetDlgItemTextA(hDlg, IDC_SETUP_HOTKEYCURRENT_GROUP, GetLangStringA(wLangSystem, STRID_HOTKEYCURRENT));
    SetDlgItemTextA(hDlg, IDC_SETUP_HOTKEYCURRENT_COLLAPSE_LABEL, GetLangStringA(wLangSystem, STRID_COLLAPSE));
    SetDlgItemTextA(hDlg, IDC_SETUP_HOTKEYCURRENT_GOBEGIN_LABEL, GetLangStringA(wLangSystem, STRID_GOBEGIN));
    SetDlgItemTextA(hDlg, IDC_SETUP_HOTKEYCURRENT_GOEND_LABEL, GetLangStringA(wLangSystem, STRID_GOEND));
    SetDlgItemTextA(hDlg, IDC_SETUP_HOTKEYCURRENT_SELECT_LABEL, GetLangStringA(wLangSystem, STRID_SELECT));
    SetDlgItemTextA(hDlg, IDC_SETUP_HOTKEYALL_GROUP, GetLangStringA(wLangSystem, STRID_HOTKEYALL));
    SetDlgItemTextA(hDlg, IDC_SETUP_HOTKEYALL_COLLAPSE_LABEL, GetLangStringA(wLangSystem, STRID_COLLAPSE));
    SetDlgItemTextA(hDlg, IDC_SETUP_HOTKEYALL_NEXTLEVEL_LABEL, GetLangStringA(wLangSystem, STRID_NEXTLEVEL));
    SetDlgItemTextA(hDlg, IDC_SETUP_HOTKEYALL_PREVLEVEL_LABEL, GetLangStringA(wLangSystem, STRID_PREVLEVEL));
    SetDlgItemTextA(hDlg, IDOK, GetLangStringA(wLangSystem, STRID_OK));
    SetDlgItemTextA(hDlg, IDCANCEL, GetLangStringA(wLangSystem, STRID_CANCEL));

    if (bShowDock) SendMessage(hWndShowDock, BM_SETCHECK, BST_CHECKED, 0);
    if (bFollowCaret) SendMessage(hWndFollowCaret, BM_SETCHECK, BST_CHECKED, 0);
    if (bNoPrintCollapsed) SendMessage(hWndNoPrintCollapsed, BM_SETCHECK, BST_CHECKED, 0);

    SendMessage(hWndCurrentCollapse, HKM_SETHOTKEY, dwCurrentCollapse, 0);
    SendMessage(hWndCurrentGoBegin, HKM_SETHOTKEY, dwCurrentGoBegin, 0);
    SendMessage(hWndCurrentGoEnd, HKM_SETHOTKEY, dwCurrentGoEnd, 0);
    SendMessage(hWndCurrentSelect, HKM_SETHOTKEY, dwCurrentSelect, 0);
    SendMessage(hWndAllCollapse, HKM_SETHOTKEY, dwAllCollapse, 0);
    SendMessage(hWndAllNextLevel, HKM_SETHOTKEY, dwAllNextLevel, 0);
    SendMessage(hWndAllPrevLevel, HKM_SETHOTKEY, dwAllPrevLevel, 0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndCurrentCollapse, pfaCurrentCollapse?dwCurrentCollapse:0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndCurrentGoBegin, pfaCurrentGoBegin?dwCurrentGoBegin:0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndCurrentGoEnd, pfaCurrentGoEnd?dwCurrentGoEnd:0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndCurrentSelect, pfaCurrentSelect?dwCurrentSelect:0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndAllCollapse, pfaAllCollapse?dwAllCollapse:0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndAllNextLevel, pfaAllNextLevel?dwAllNextLevel:0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndAllPrevLevel, pfaAllPrevLevel?dwAllPrevLevel:0);

    crPanelFirstBackgroundDlg=crPanelFirstBackground;
    crPanelSecondBackgroundDlg=crPanelSecondBackground;
    crPanelNormalFoldDlg=crPanelNormalFold;
    crPanelActiveFoldDlg=crPanelActiveFold;
  }
  else if (uMsg == WM_DRAWITEM)
  {
    DRAWITEMSTRUCT *dis=(DRAWITEMSTRUCT *)lParam;
    HBRUSH hBrush;
    DWORD dwColor=GetSysColor(COLOR_BTNFACE);

    if (wParam == IDC_SETUP_FIRSTBACKGROUNDCOLOR_BUTTON)
      dwColor=crPanelFirstBackgroundDlg;
    else if (wParam == IDC_SETUP_SECONDBACKGROUNDCOLOR_BUTTON)
      dwColor=crPanelSecondBackgroundDlg;
    else if (wParam == IDC_SETUP_NORMALFOLDCOLOR_BUTTON)
      dwColor=crPanelNormalFoldDlg;
    else if (wParam == IDC_SETUP_ACTIVEFOLDCOLOR_BUTTON)
      dwColor=crPanelActiveFoldDlg;

    if (hBrush=CreateSolidBrush(dwColor))
    {
      FillRect(dis->hDC, &dis->rcItem, hBrush);
      if (dis->hwndItem == GetFocus())
        DrawFocusRect(dis->hDC, &dis->rcItem);
      DeleteObject(hBrush);
    }
    return TRUE;
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_SETUP_FIRSTBACKGROUNDCOLOR_BUTTON ||
        LOWORD(wParam) == IDC_SETUP_SECONDBACKGROUNDCOLOR_BUTTON ||
        LOWORD(wParam) == IDC_SETUP_NORMALFOLDCOLOR_BUTTON ||
        LOWORD(wParam) == IDC_SETUP_ACTIVEFOLDCOLOR_BUTTON)
    {
      CHOOSECOLORA ccA={0};
      COLORREF crCustColors[16]={0};
      COLORREF *lpcrColor=NULL;
      HWND hWndButton=NULL;

      if (LOWORD(wParam) == IDC_SETUP_FIRSTBACKGROUNDCOLOR_BUTTON)
      {
        lpcrColor=&crPanelFirstBackgroundDlg;
        hWndButton=hWndFirstBackgroundColor;
      }
      else if (LOWORD(wParam) == IDC_SETUP_SECONDBACKGROUNDCOLOR_BUTTON)
      {
        lpcrColor=&crPanelSecondBackgroundDlg;
        hWndButton=hWndSecondBackgroundColor;
      }
      else if (LOWORD(wParam) == IDC_SETUP_NORMALFOLDCOLOR_BUTTON)
      {
        lpcrColor=&crPanelNormalFoldDlg;
        hWndButton=hWndNormalFoldColor;
      }
      else if (LOWORD(wParam) == IDC_SETUP_ACTIVEFOLDCOLOR_BUTTON)
      {
        lpcrColor=&crPanelActiveFoldDlg;
        hWndButton=hWndActiveFoldColor;
      }
      ccA.lStructSize  =sizeof(CHOOSECOLORA);
      ccA.hwndOwner    =hDlg;
      ccA.lpCustColors =crCustColors;
      ccA.Flags        =CC_FULLOPEN|CC_RGBINIT;
      ccA.rgbResult    =*lpcrColor;

      if (ChooseColorA(&ccA))
      {
        *lpcrColor=ccA.rgbResult;
        InvalidateRect(hWndButton, NULL, FALSE);
      }
    }
    else if (LOWORD(wParam) == IDOK)
    {
      BOOL bState;

      crPanelFirstBackground=crPanelFirstBackgroundDlg;
      crPanelSecondBackground=crPanelSecondBackgroundDlg;
      crPanelNormalFold=crPanelNormalFoldDlg;
      crPanelActiveFold=crPanelActiveFoldDlg;

      bState=SendMessage(hWndShowDock, BM_GETCHECK, 0, 0);
      if (bState != bShowDock)
      {
        bShowDock=bState;

        if (bShowDock)
          CreateDock(&hWndMainDlg, &dkMainDlg);
        else
          DestroyDock(hWndMainDlg, DKT_NOUNLOAD);
      }
      bFollowCaret=SendMessage(hWndFollowCaret, BM_GETCHECK, 0, 0);
      bNoPrintCollapsed=SendMessage(hWndNoPrintCollapsed, BM_GETCHECK, 0, 0);

      dwCurrentCollapse=(WORD)SendMessage(hWndCurrentCollapse, HKM_GETHOTKEY, 0, 0);
      dwCurrentGoBegin=(WORD)SendMessage(hWndCurrentGoBegin, HKM_GETHOTKEY, 0, 0);
      dwCurrentGoEnd=(WORD)SendMessage(hWndCurrentGoEnd, HKM_GETHOTKEY, 0, 0);
      dwCurrentSelect=(WORD)SendMessage(hWndCurrentSelect, HKM_GETHOTKEY, 0, 0);
      dwAllCollapse=(WORD)SendMessage(hWndAllCollapse, HKM_GETHOTKEY, 0, 0);
      dwAllNextLevel=(WORD)SendMessage(hWndAllNextLevel, HKM_GETHOTKEY, 0, 0);
      dwAllPrevLevel=(WORD)SendMessage(hWndAllPrevLevel, HKM_GETHOTKEY, 0, 0);

      SaveOptionsA(OF_SETTINGS);
      EndDialog(hDlg, 0);

      if (bInitMain)
      {
        if (hWndEdit=GetCurEdit())
        {
          StackBeginBoards();
          lpCurrentFoldWindow=SetActiveEdit(hWndEdit, hWndFunctionList, SAE_RESETLIST);
        }
      }
      return TRUE;
    }
    else if (LOWORD(wParam) == IDCANCEL)
    {
      EndDialog(hDlg, 0);
      return TRUE;
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }
  return FALSE;
}

BOOL CALLBACK SetupDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndShowDock;
  static HWND hWndFollowCaret;
  static HWND hWndNoPrintCollapsed;
  static HWND hWndFirstBackgroundColor;
  static HWND hWndSecondBackgroundColor;
  static HWND hWndNormalFoldColor;
  static HWND hWndActiveFoldColor;
  static HWND hWndCurrentCollapse;
  static HWND hWndCurrentGoBegin;
  static HWND hWndCurrentGoEnd;
  static HWND hWndCurrentSelect;
  static HWND hWndAllCollapse;
  static HWND hWndAllNextLevel;
  static HWND hWndAllPrevLevel;
  static COLORREF crPanelFirstBackgroundDlg;
  static COLORREF crPanelSecondBackgroundDlg;
  static COLORREF crPanelNormalFoldDlg;
  static COLORREF crPanelActiveFoldDlg;

  if (uMsg == WM_INITDIALOG)
  {
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hMainIcon);
    hWndShowDock=GetDlgItem(hDlg, IDC_SETUP_SHOWDOCK);
    hWndFollowCaret=GetDlgItem(hDlg, IDC_SETUP_FOLLOWCARET);
    hWndNoPrintCollapsed=GetDlgItem(hDlg, IDC_SETUP_NOPRINTCOLLAPSED);
    hWndFirstBackgroundColor=GetDlgItem(hDlg, IDC_SETUP_FIRSTBACKGROUNDCOLOR_BUTTON);
    hWndSecondBackgroundColor=GetDlgItem(hDlg, IDC_SETUP_SECONDBACKGROUNDCOLOR_BUTTON);
    hWndNormalFoldColor=GetDlgItem(hDlg, IDC_SETUP_NORMALFOLDCOLOR_BUTTON);
    hWndActiveFoldColor=GetDlgItem(hDlg, IDC_SETUP_ACTIVEFOLDCOLOR_BUTTON);
    hWndCurrentCollapse=GetDlgItem(hDlg, IDC_SETUP_HOTKEYCURRENT_COLLAPSE);
    hWndCurrentGoBegin=GetDlgItem(hDlg, IDC_SETUP_HOTKEYCURRENT_GOBEGIN);
    hWndCurrentGoEnd=GetDlgItem(hDlg, IDC_SETUP_HOTKEYCURRENT_GOEND);
    hWndCurrentSelect=GetDlgItem(hDlg, IDC_SETUP_HOTKEYCURRENT_SELECT);
    hWndAllCollapse=GetDlgItem(hDlg, IDC_SETUP_HOTKEYALL_COLLAPSE);
    hWndAllNextLevel=GetDlgItem(hDlg, IDC_SETUP_HOTKEYALL_NEXTLEVEL);
    hWndAllPrevLevel=GetDlgItem(hDlg, IDC_SETUP_HOTKEYALL_PREVLEVEL);

    SetDlgItemTextW(hDlg, IDC_SETUP_DOCK_GROUP, GetLangStringW(wLangSystem, STRID_DOCKWINDOW));
    SetDlgItemTextW(hDlg, IDC_SETUP_SHOWDOCK, GetLangStringW(wLangSystem, STRID_SHOWDOCK));
    SetDlgItemTextW(hDlg, IDC_SETUP_FOLLOWCARET, GetLangStringW(wLangSystem, STRID_FOLLOWCARET));
    SetDlgItemTextW(hDlg, IDC_SETUP_OTHER_GROUP, GetLangStringW(wLangSystem, STRID_OTHERSETTINGS));
    SetDlgItemTextW(hDlg, IDC_SETUP_NOPRINTCOLLAPSED, GetLangStringW(wLangSystem, STRID_NOPRINTCOLLAPSED));
    SetDlgItemTextW(hDlg, IDC_SETUP_COLORS_GROUP, GetLangStringW(wLangSystem, STRID_COLORS));
    SetDlgItemTextW(hDlg, IDC_SETUP_FIRSTBACKGROUNDCOLOR_LABEL, GetLangStringW(wLangSystem, STRID_FIRSTBACKGROUNDCOLOR));
    SetDlgItemTextW(hDlg, IDC_SETUP_SECONDBACKGROUNDCOLOR_LABEL, GetLangStringW(wLangSystem, STRID_SECONDBACKGROUNDCOLOR));
    SetDlgItemTextW(hDlg, IDC_SETUP_NORMALFOLDCOLOR_LABEL, GetLangStringW(wLangSystem, STRID_NORMALFOLDCOLOR));
    SetDlgItemTextW(hDlg, IDC_SETUP_ACTIVEFOLDCOLOR_LABEL, GetLangStringW(wLangSystem, STRID_ACTIVEFOLDCOLOR));
    SetDlgItemTextW(hDlg, IDC_SETUP_HOTKEYCURRENT_GROUP, GetLangStringW(wLangSystem, STRID_HOTKEYCURRENT));
    SetDlgItemTextW(hDlg, IDC_SETUP_HOTKEYCURRENT_COLLAPSE_LABEL, GetLangStringW(wLangSystem, STRID_COLLAPSE));
    SetDlgItemTextW(hDlg, IDC_SETUP_HOTKEYCURRENT_GOBEGIN_LABEL, GetLangStringW(wLangSystem, STRID_GOBEGIN));
    SetDlgItemTextW(hDlg, IDC_SETUP_HOTKEYCURRENT_GOEND_LABEL, GetLangStringW(wLangSystem, STRID_GOEND));
    SetDlgItemTextW(hDlg, IDC_SETUP_HOTKEYCURRENT_SELECT_LABEL, GetLangStringW(wLangSystem, STRID_SELECT));
    SetDlgItemTextW(hDlg, IDC_SETUP_HOTKEYALL_GROUP, GetLangStringW(wLangSystem, STRID_HOTKEYALL));
    SetDlgItemTextW(hDlg, IDC_SETUP_HOTKEYALL_COLLAPSE_LABEL, GetLangStringW(wLangSystem, STRID_COLLAPSE));
    SetDlgItemTextW(hDlg, IDC_SETUP_HOTKEYALL_NEXTLEVEL_LABEL, GetLangStringW(wLangSystem, STRID_NEXTLEVEL));
    SetDlgItemTextW(hDlg, IDC_SETUP_HOTKEYALL_PREVLEVEL_LABEL, GetLangStringW(wLangSystem, STRID_PREVLEVEL));
    SetDlgItemTextW(hDlg, IDOK, GetLangStringW(wLangSystem, STRID_OK));
    SetDlgItemTextW(hDlg, IDCANCEL, GetLangStringW(wLangSystem, STRID_CANCEL));

    if (bShowDock) SendMessage(hWndShowDock, BM_SETCHECK, BST_CHECKED, 0);
    if (bFollowCaret) SendMessage(hWndFollowCaret, BM_SETCHECK, BST_CHECKED, 0);
    if (bNoPrintCollapsed) SendMessage(hWndNoPrintCollapsed, BM_SETCHECK, BST_CHECKED, 0);

    SendMessage(hWndCurrentCollapse, HKM_SETHOTKEY, dwCurrentCollapse, 0);
    SendMessage(hWndCurrentGoBegin, HKM_SETHOTKEY, dwCurrentGoBegin, 0);
    SendMessage(hWndCurrentGoEnd, HKM_SETHOTKEY, dwCurrentGoEnd, 0);
    SendMessage(hWndCurrentSelect, HKM_SETHOTKEY, dwCurrentSelect, 0);
    SendMessage(hWndAllCollapse, HKM_SETHOTKEY, dwAllCollapse, 0);
    SendMessage(hWndAllNextLevel, HKM_SETHOTKEY, dwAllNextLevel, 0);
    SendMessage(hWndAllPrevLevel, HKM_SETHOTKEY, dwAllPrevLevel, 0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndCurrentCollapse, pfwCurrentCollapse?dwCurrentCollapse:0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndCurrentGoBegin, pfwCurrentGoBegin?dwCurrentGoBegin:0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndCurrentGoEnd, pfwCurrentGoEnd?dwCurrentGoEnd:0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndCurrentSelect, pfwCurrentSelect?dwCurrentSelect:0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndAllCollapse, pfwAllCollapse?dwAllCollapse:0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndAllNextLevel, pfwAllNextLevel?dwAllNextLevel:0);
    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndAllPrevLevel, pfwAllPrevLevel?dwAllPrevLevel:0);

    crPanelFirstBackgroundDlg=crPanelFirstBackground;
    crPanelSecondBackgroundDlg=crPanelSecondBackground;
    crPanelNormalFoldDlg=crPanelNormalFold;
    crPanelActiveFoldDlg=crPanelActiveFold;
  }
  else if (uMsg == WM_DRAWITEM)
  {
    DRAWITEMSTRUCT *dis=(DRAWITEMSTRUCT *)lParam;
    HBRUSH hBrush;
    DWORD dwColor=GetSysColor(COLOR_BTNFACE);

    if (wParam == IDC_SETUP_FIRSTBACKGROUNDCOLOR_BUTTON)
      dwColor=crPanelFirstBackgroundDlg;
    else if (wParam == IDC_SETUP_SECONDBACKGROUNDCOLOR_BUTTON)
      dwColor=crPanelSecondBackgroundDlg;
    else if (wParam == IDC_SETUP_NORMALFOLDCOLOR_BUTTON)
      dwColor=crPanelNormalFoldDlg;
    else if (wParam == IDC_SETUP_ACTIVEFOLDCOLOR_BUTTON)
      dwColor=crPanelActiveFoldDlg;

    if (hBrush=CreateSolidBrush(dwColor))
    {
      FillRect(dis->hDC, &dis->rcItem, hBrush);
      if (dis->hwndItem == GetFocus())
        DrawFocusRect(dis->hDC, &dis->rcItem);
      DeleteObject(hBrush);
    }
    return TRUE;
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_SETUP_FIRSTBACKGROUNDCOLOR_BUTTON ||
        LOWORD(wParam) == IDC_SETUP_SECONDBACKGROUNDCOLOR_BUTTON ||
        LOWORD(wParam) == IDC_SETUP_NORMALFOLDCOLOR_BUTTON ||
        LOWORD(wParam) == IDC_SETUP_ACTIVEFOLDCOLOR_BUTTON)
    {
      CHOOSECOLORW ccW={0};
      COLORREF crCustColors[16]={0};
      COLORREF *lpcrColor=NULL;
      HWND hWndButton=NULL;

      if (LOWORD(wParam) == IDC_SETUP_FIRSTBACKGROUNDCOLOR_BUTTON)
      {
        lpcrColor=&crPanelFirstBackgroundDlg;
        hWndButton=hWndFirstBackgroundColor;
      }
      else if (LOWORD(wParam) == IDC_SETUP_SECONDBACKGROUNDCOLOR_BUTTON)
      {
        lpcrColor=&crPanelSecondBackgroundDlg;
        hWndButton=hWndSecondBackgroundColor;
      }
      else if (LOWORD(wParam) == IDC_SETUP_NORMALFOLDCOLOR_BUTTON)
      {
        lpcrColor=&crPanelNormalFoldDlg;
        hWndButton=hWndNormalFoldColor;
      }
      else if (LOWORD(wParam) == IDC_SETUP_ACTIVEFOLDCOLOR_BUTTON)
      {
        lpcrColor=&crPanelActiveFoldDlg;
        hWndButton=hWndActiveFoldColor;
      }
      ccW.lStructSize  =sizeof(CHOOSECOLORW);
      ccW.hwndOwner    =hDlg;
      ccW.lpCustColors =crCustColors;
      ccW.Flags        =CC_FULLOPEN|CC_RGBINIT;
      ccW.rgbResult    =*lpcrColor;

      if (ChooseColorW(&ccW))
      {
        *lpcrColor=ccW.rgbResult;
        InvalidateRect(hWndButton, NULL, FALSE);
      }
    }
    else if (LOWORD(wParam) == IDOK)
    {
      BOOL bState;

      crPanelFirstBackground=crPanelFirstBackgroundDlg;
      crPanelSecondBackground=crPanelSecondBackgroundDlg;
      crPanelNormalFold=crPanelNormalFoldDlg;
      crPanelActiveFold=crPanelActiveFoldDlg;

      bState=SendMessage(hWndShowDock, BM_GETCHECK, 0, 0);
      if (bState != bShowDock)
      {
        bShowDock=bState;

        if (bShowDock)
          CreateDock(&hWndMainDlg, &dkMainDlg);
        else
          DestroyDock(hWndMainDlg, DKT_NOUNLOAD);
      }
      bFollowCaret=SendMessage(hWndFollowCaret, BM_GETCHECK, 0, 0);
      bNoPrintCollapsed=SendMessage(hWndNoPrintCollapsed, BM_GETCHECK, 0, 0);

      dwCurrentCollapse=(WORD)SendMessage(hWndCurrentCollapse, HKM_GETHOTKEY, 0, 0);
      dwCurrentGoBegin=(WORD)SendMessage(hWndCurrentGoBegin, HKM_GETHOTKEY, 0, 0);
      dwCurrentGoEnd=(WORD)SendMessage(hWndCurrentGoEnd, HKM_GETHOTKEY, 0, 0);
      dwCurrentSelect=(WORD)SendMessage(hWndCurrentSelect, HKM_GETHOTKEY, 0, 0);
      dwAllCollapse=(WORD)SendMessage(hWndAllCollapse, HKM_GETHOTKEY, 0, 0);
      dwAllNextLevel=(WORD)SendMessage(hWndAllNextLevel, HKM_GETHOTKEY, 0, 0);
      dwAllPrevLevel=(WORD)SendMessage(hWndAllPrevLevel, HKM_GETHOTKEY, 0, 0);

      SaveOptionsW(OF_SETTINGS);
      EndDialog(hDlg, 0);

      if (bInitMain)
      {
        if (hWndEdit=GetCurEdit())
        {
          StackBeginBoards();
          lpCurrentFoldWindow=SetActiveEdit(hWndEdit, hWndFunctionList, SAE_RESETLIST);
        }
      }
      return TRUE;
    }
    else if (LOWORD(wParam) == IDCANCEL)
    {
      EndDialog(hDlg, 0);
      return TRUE;
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }
  return FALSE;
}

LRESULT CALLBACK NewMainProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  LRESULT lResult;

  if (uMsg == AKDN_SIZE)
  {
    if (bDockWaitResize)
    {
      bDockWaitResize=FALSE;
      ShowWindow(hWndMainDlg, SW_SHOW);

      if (hWndEdit=GetCurEdit())
      {
        StackBeginBoards();
        lpCurrentFoldWindow=SetActiveEdit(hWndEdit, hWndFunctionList, SAE_RESETLIST);
      }
    }
  }
  else if (uMsg == AKDN_EDIT_ONFINISH)
  {
    FOLDWINDOW *lpFoldWindow;
    MANUALSET *lpManual;

    if (lpManual=StackGetManual(&hManualStack, (HWND)wParam))
      StackDeleteManual(&hManualStack, lpManual);
    if (lpFoldWindow=StackGetFoldWindow(&hWindowsStack, (HWND)wParam))
      StackDeleteFoldWindow(&hWindowsStack, lpFoldWindow);
  }
  else if (uMsg == AKDN_MAIN_ONFINISH)
  {
    NewMainProcData->NextProc(hWnd, uMsg, wParam, lParam);

    if (hWndMainDlg)
    {
      DestroyDock(hWndMainDlg, DKT_ONMAINFINISH);
    }
    else
    {
      StackFreeFoldWindows(&hWindowsStack);
      UninitMain();
    }
    return FALSE;
  }
  if (lResult=EditParentMessages(hWnd, uMsg, wParam, lParam))
    return lResult;

  //Call next procedure
  return NewMainProcData->NextProc(hWnd, uMsg, wParam, lParam);
}

LRESULT CALLBACK NewFrameProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  LRESULT lResult;

  if (lResult=EditParentMessages(hWnd, uMsg, wParam, lParam))
    return lResult;

  //Call next procedure
  return NewFrameProcData->NextProc(hWnd, uMsg, wParam, lParam);
}

LRESULT CALLBACK EditParentMessages(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static BOOL bResetFold=FALSE;

  if (uMsg == AKDN_FRAME_NOWINDOWS)
  {
    SendMessage(hWndFunctionList, WM_SETREDRAW, FALSE, 0);
    SendMessage(hWndFunctionList, TVM_DELETEITEM, 0, (LPARAM)NULL);
    SendMessage(hWndFunctionList, WM_SETREDRAW, TRUE, 0);
    InvalidateRect(hWndFunctionList, NULL, TRUE);
  }
  else if (uMsg == AKDN_OPENDOCUMENT_START)
  {
    bResetFold=TRUE;
  }
  else if (uMsg == AKDN_SAVEDOCUMENT_START)
  {
    NSAVEDOCUMENTA *nsdA=(NSAVEDOCUMENTA *)lParam;
    NSAVEDOCUMENTW *nsdW=(NSAVEDOCUMENTW *)lParam;
    EDITINFO ei;

    if (SendMessage(hMainWnd, AKD_GETEDITINFO, wParam, (LPARAM)&ei))
    {
      if (bOldWindows)
      {
        if (lstrcmpiA(GetFileExtA(nsdA->pFile), GetFileExtA((char *)ei.pFile)))
          bResetFold=TRUE;
      }
      else
      {
        if (lstrcmpiW(GetFileExtW(nsdW->pFile), GetFileExtW((wchar_t *)ei.pFile)))
          bResetFold=TRUE;
      }
    }
  }
  else if (uMsg == AKDN_OPENDOCUMENT_FINISH ||
           uMsg == AKDN_SAVEDOCUMENT_FINISH)
  {
    if (bResetFold)
    {
      bResetFold=FALSE;

      if (lpCurrentFoldWindow && lpCurrentFoldWindow->hWnd == (HWND)wParam)
        SetActiveEdit((HWND)wParam, hWndFunctionList, SAE_RESETFOLD|SAE_RESETLIST);
      else
        SetActiveEdit((HWND)wParam, hWndFunctionList, SAE_RESETFOLD);
    }
  }
  else if (uMsg == WM_NOTIFY)
  {
    if (wParam == ID_EDIT)
    {
      if (((NMHDR *)lParam)->code == AEN_SETFOCUS)
      {
        if (!lpCurrentFoldWindow || lpCurrentFoldWindow->hWnd != ((NMHDR *)lParam)->hwndFrom)
        {
          lpCurrentFoldWindow=SetActiveEdit(((NMHDR *)lParam)->hwndFrom, hWndFunctionList, SAE_RESETLIST);
        }
      }
      else if (((NMHDR *)lParam)->code == AEN_TEXTCHANGING)
      {
        AENTEXTCHANGE *aentc=(AENTEXTCHANGE *)lParam;
        FOLDWINDOW *lpFoldWindow;
        LEVEL *lpMinLevel;
        LEVEL *lpMaxLevel;

        if (lpFoldWindow=StackGetFoldWindow(&hWindowsStack, aentc->hdr.hwndFrom))
        {
          if (lpFoldWindow->lpSyntaxFile)
          {
            lpFoldWindow->bLevelChanged=FALSE;
            lpFoldWindow->lpTextChangeMinParent=NULL;
            lpFoldWindow->lpTextChangeMaxParent=NULL;
            memset(&lpFoldWindow->crTextChange, 0, sizeof(AECHARRANGE));

            //Get min parent level.
            lpMinLevel=NULL;
            if (StackGetLevelByIndex(&lpFoldWindow->hLevelStack, &lpMinLevel, &aentc->aes.crSel.ciMin, FALSE))
            {
              for (lpFoldWindow->lpTextChangeMinParent=lpMinLevel; lpFoldWindow->lpTextChangeMinParent->lpParent; lpFoldWindow->lpTextChangeMinParent=lpFoldWindow->lpTextChangeMinParent->lpParent);
            }

            //Get max parent level.
            lpMaxLevel=lpMinLevel;
            if (StackGetLevelByIndex(&lpFoldWindow->hLevelStack, &lpMaxLevel, &aentc->aes.crSel.ciMax, FALSE))
            {
              for (lpFoldWindow->lpTextChangeMaxParent=lpMaxLevel; lpFoldWindow->lpTextChangeMaxParent->lpParent; lpFoldWindow->lpTextChangeMaxParent=lpFoldWindow->lpTextChangeMaxParent->lpParent);
            }
          }
        }
      }
      else if (((NMHDR *)lParam)->code == AEN_TEXTINSERTEND ||
               ((NMHDR *)lParam)->code == AEN_TEXTDELETEEND)
      {
        FOLDWINDOW *lpFoldWindow;
        AECHARRANGE *lpcrRange;

        if (lpFoldWindow=StackGetFoldWindow(&hWindowsStack, ((NMHDR *)lParam)->hwndFrom))
        {
          if (lpFoldWindow->lpSyntaxFile)
          {
            if (((NMHDR *)lParam)->code == AEN_TEXTINSERTEND)
              lpcrRange=&((AENTEXTINSERT *)lParam)->crAkelRange;
            else
              lpcrRange=&((AENTEXTDELETE *)lParam)->crAkelRange;

            if (lpcrRange->ciMin.nLine == lpFoldWindow->crTextChange.ciMin.nLine)
              lpFoldWindow->crTextChange.ciMin.lpLine=lpcrRange->ciMin.lpLine;
            if (lpcrRange->ciMax.nLine == lpFoldWindow->crTextChange.ciMax.nLine)
              lpFoldWindow->crTextChange.ciMax.lpLine=lpcrRange->ciMax.lpLine;

            if (!lpFoldWindow->crTextChange.ciMin.lpLine || AEC_IndexCompare(&lpcrRange->ciMin, &lpFoldWindow->crTextChange.ciMin) <= 0)
              lpFoldWindow->crTextChange.ciMin=lpcrRange->ciMin;
            if (!lpFoldWindow->crTextChange.ciMax.lpLine || AEC_IndexCompare(&lpcrRange->ciMax, &lpFoldWindow->crTextChange.ciMax) >= 0)
              lpFoldWindow->crTextChange.ciMax=lpcrRange->ciMax;
          }
        }
      }
      else if (((NMHDR *)lParam)->code == AEN_POINT)
      {
        AENPOINT *aenpt=(AENPOINT *)lParam;
        FOLDWINDOW *lpFoldWindow;

        if ((aenpt->dwType & AEPTT_SETTEXT) ||
            (aenpt->dwType & AEPTT_STREAMIN) ||
            (aenpt->lpPoint->dwFlags & AEPTF_FOLD))
        {
          if (lpFoldWindow=StackGetFoldWindow(&hWindowsStack, aenpt->hdr.hwndFrom))
          {
            if (lpFoldWindow->lpSyntaxFile)
            {
              lpFoldWindow->bLevelChanged=TRUE;
            }
          }
        }
      }
      else if (((NMHDR *)lParam)->code == AEN_TEXTCHANGED)
      {
        AENTEXTCHANGE *aentc=(AENTEXTCHANGE *)lParam;
        FOLDWINDOW *lpFoldWindow;
        LEVEL *lpLevelExist;
        LEVEL *lpLevelTemp;
        LEVEL *lpLevelNextExist;
        LEVEL *lpLevelNextTemp;
        int nCompare;

        if (lpFoldWindow=StackGetFoldWindow(&hWindowsStack, aentc->hdr.hwndFrom))
        {
          if (lpFoldWindow->lpSyntaxFile)
          {
            if (aentc->dwType & AETCT_DELETEALL)
            {
              //All text is changed.
              if (lpCurrentFoldWindow && lpCurrentFoldWindow->hWnd == lpFoldWindow->hWnd)
                SetActiveEdit(lpFoldWindow->hWnd, hWndFunctionList, SAE_RESETFOLD|SAE_RESETLIST);
              else
                SetActiveEdit(lpFoldWindow->hWnd, hWndFunctionList, SAE_RESETFOLD);
            }
            else if (!(aentc->dwType & AETCT_NONE))
            {
              if (!lpFoldWindow->bLevelChanged)
              {
                AECHARRANGE crRange;
                LEVEL *lpMinLevel;

                if (lpFoldWindow->crTextChange.ciMin.lpLine &&
                    lpFoldWindow->crTextChange.ciMax.lpLine)
                {
                  crRange=lpFoldWindow->crTextChange;
                  SendMessage(lpFoldWindow->hWnd, AEM_GETINDEX, AEGI_WRAPLINEBEGIN, (LPARAM)&crRange.ciMin);
                  SendMessage(lpFoldWindow->hWnd, AEM_GETINDEX, AEGI_WRAPLINEEND, (LPARAM)&crRange.ciMax);
                  lpMinLevel=lpFoldWindow->lpTextChangeMinParent;
                  StackGetLevelByLine(&lpFoldWindow->hLevelStack, &lpMinLevel, crRange.ciMin.nLine, FALSE);

                  //Find is any new folds appeared or existed doesn't match.
                  if (FindFold(lpFoldWindow, lpMinLevel, &crRange))
                    lpFoldWindow->bLevelChanged=TRUE;
                }
              }

              //Levels were changed, update info.
              if (lpFoldWindow->bLevelChanged)
              {
                //Get start index for rescan levels.
                if (!lpFoldWindow->lpTextChangeMinParent)
                {
                  lpFoldWindow->lpTextChangeMinParent=(LEVEL *)lpFoldWindow->hLevelStack.first;
                  SendMessage(lpFoldWindow->hWnd, AEM_GETINDEX, AEGI_FIRSTCHAR, (LPARAM)&lpFoldWindow->crTextChange.ciMin);
                }
                else lpFoldWindow->crTextChange.ciMin=lpFoldWindow->lpTextChangeMinParent->lpFold->lpMinPoint->ciPoint;

                //Get end index for rescan levels.
                if (!lpFoldWindow->lpTextChangeMaxParent)
                {
                  lpFoldWindow->lpTextChangeMaxParent=(LEVEL *)lpFoldWindow->hLevelStack.last;
                  SendMessage(lpFoldWindow->hWnd, AEM_GETINDEX, AEGI_LASTCHAR, (LPARAM)&lpFoldWindow->crTextChange.ciMax);
                }
                else lpFoldWindow->crTextChange.ciMax=lpFoldWindow->lpTextChangeMaxParent->lpFold->lpMaxPoint->ciPoint;

                //Parse changed text range and put results in temp stack.
                lpFoldWindow->hTempStack.first=0;
                lpFoldWindow->hTempStack.last=0;
                FillFoldsStack(lpFoldWindow, &lpFoldWindow->hTempStack, lpFoldWindow->hWnd, &lpFoldWindow->crTextChange);

                //Transfer existed elements settings that matched with new ones to temp stack and delete them.
                lpLevelExist=lpFoldWindow->lpTextChangeMinParent;
                lpLevelTemp=(LEVEL *)lpFoldWindow->hTempStack.first;

                while (lpLevelExist && lpLevelTemp)
                {
                  lpLevelNextExist=lpLevelExist->next;
                  lpLevelNextTemp=lpLevelTemp->next;

                  nCompare=AEC_IndexCompare(&lpLevelExist->lpFold->lpMinPoint->ciPoint, &lpLevelTemp->pointMin.ciPoint);

                  if (nCompare <= 0)
                  {
                    if (nCompare == 0)
                    {
                      if (lpLevelExist->lpFoldInfo == lpLevelTemp->lpFoldInfo)
                      {
                        if (!AEC_IndexCompare(&lpLevelExist->lpFold->lpMaxPoint->ciPoint, &lpLevelTemp->pointMax.ciPoint))
                        {
                          if (!(lpLevelExist->lpFold->lpMinPoint->dwFlags & AEPTF_MODIFY) &&
                              !(lpLevelExist->lpFold->lpMaxPoint->dwFlags & AEPTF_MODIFY))
                          {
                            lpLevelTemp->lpFold=lpLevelExist->lpFold;
                            lpLevelExist->lpFold=NULL;
                          }
                        }
                      }
                      lpLevelTemp=lpLevelNextTemp;
                    }
                    StackDeleteLevel(&lpFoldWindow->hLevelStack, lpFoldWindow->hWnd, lpLevelExist);
                    lpLevelExist=lpLevelNextExist;
                  }
                  else lpLevelTemp=lpLevelNextTemp;
                }

                //Delete remaining existed elements that are in changed text area.
                while (lpLevelExist)
                {
                  lpLevelNextExist=lpLevelExist->next;

                  if (AEC_IndexCompare(&lpLevelExist->lpFold->lpMinPoint->ciPoint, &lpFoldWindow->crTextChange.ciMax) <= 0)
                    StackDeleteLevel(&lpFoldWindow->hLevelStack, lpFoldWindow->hWnd, lpLevelExist);
                  else
                    break;

                  lpLevelExist=lpLevelNextExist;
                }

                //Fill lpFold member and insert temp stack to an existent.
                AddFoldsToStack(&lpFoldWindow->hTempStack, lpFoldWindow->hWnd);
                StackJoin((stack **)&lpFoldWindow->hLevelStack.first, (stack **)&lpFoldWindow->hLevelStack.last, (stack *)lpLevelExist, (stack *)lpFoldWindow->hTempStack.first, (stack *)lpFoldWindow->hTempStack.last);

                //Update list and edit control
                SetActiveEdit(lpFoldWindow->hWnd, hWndFunctionList, SAE_RESETLIST);
                InvalidateRect(lpFoldWindow->hWnd, NULL, TRUE);
              }
              lpFoldWindow->bLevelChanged=FALSE;
              memset(&lpFoldWindow->crTextChange, 0, sizeof(AECHARRANGE));
            }
          }
        }
      }
      else if (((NMHDR *)lParam)->code == AEN_SELCHANGED)
      {
        AENSELCHANGE *aensc=(AENSELCHANGE *)lParam;

        if (bShowDock)
        {
          if (lpCurrentFoldWindow && lpCurrentFoldWindow->hWnd == aensc->hdr.hwndFrom)
          {
            if (!lpCurrentFoldWindow->bLockSel)
              SetCaretTreeView(hWndFunctionList, lpCurrentFoldWindow);
          }
        }
        SendMessage(aensc->hdr.hwndFrom, AEM_COLLAPSELINE, aensc->ciCaret.nLine, FALSE);
      }
      else if (((NMHDR *)lParam)->code == AEN_SETRECT)
      {
        AENSETRECT *aensr=(AENSETRECT *)lParam;
        FOLDWINDOW *lpFoldWindow;
        RECT rcLineBoard={0};

        if (lpFoldWindow=StackGetFoldWindow(&hWindowsStack, aensr->hdr.hwndFrom))
        {
          if (lpFoldWindow->rcBoard.right)
          {
            if (bOldWindows)
            {
              PLUGINFUNCTIONA *pf;

              if (pf=(PLUGINFUNCTIONA *)SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)"LineBoard::Main", 0))
              {
                if (pf->bRunning)
                {
                  PLUGINCALLSENDA pcs;
                  DLLECLINEBOARD declb;

                  declb.dwStructSize=sizeof(DLLECLINEBOARD);
                  declb.nAction=DLLA_LINEBOARD_GETBOARDRECT;
                  declb.hWndEdit=aensr->hdr.hwndFrom;
                  declb.rcBoard=&rcLineBoard;

                  pcs.pFunction="LineBoard::Main";
                  pcs.bOnStart=FALSE;
                  pcs.lParam=(LPARAM)&declb;
                  pcs.lpbAutoLoad=NULL;
                  SendMessage(hMainWnd, AKD_DLLCALL, 0, (LPARAM)&pcs);
                }
              }
            }
            else
            {
              PLUGINFUNCTIONW *pf;

              if (pf=(PLUGINFUNCTIONW *)SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)L"LineBoard::Main", 0))
              {
                if (pf->bRunning)
                {
                  PLUGINCALLSENDW pcs;
                  DLLECLINEBOARD declb;

                  declb.dwStructSize=sizeof(DLLECLINEBOARD);
                  declb.nAction=DLLA_LINEBOARD_GETBOARDRECT;
                  declb.hWndEdit=aensr->hdr.hwndFrom;
                  declb.rcBoard=&rcLineBoard;

                  pcs.pFunction=L"LineBoard::Main";
                  pcs.bOnStart=FALSE;
                  pcs.lParam=(LPARAM)&declb;
                  pcs.lpbAutoLoad=NULL;
                  SendMessage(hMainWnd, AKD_DLLCALL, 0, (LPARAM)&pcs);
                }
              }
            }
            lpFoldWindow->rcBoard.right=rcLineBoard.right + (lpFoldWindow->rcBoard.right - lpFoldWindow->rcBoard.left);
            lpFoldWindow->rcBoard.left=rcLineBoard.right;
          }
        }
      }
    }
  }
  return FALSE;
}

LRESULT CALLBACK NewEditProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == WM_ERASEBKGND)
  {
    FOLDWINDOW *lpFoldWindow;

    if (lpFoldWindow=StackGetFoldWindow(&hWindowsStack, hWnd))
    {
      if (lpFoldWindow->lpSyntaxFile)
      {
        SendMessage(hWnd, AEM_LOCKERASERECT, 0, (LPARAM)&lpFoldWindow->rcBoard);
      }
    }
  }
  else if (uMsg == WM_SIZE)
  {
    FOLDWINDOW *lpFoldWindow;

    if (lpFoldWindow=StackGetFoldWindow(&hWindowsStack, hWnd))
    {
      if (lpFoldWindow->lpSyntaxFile)
      {
        lpFoldWindow->rcBoard.bottom=HIWORD(lParam);
      }
    }
  }
  else if (uMsg == AEM_STARTPRINTDOC)
  {
    FOLDWINDOW *lpFoldWindow;
    DWORD dwOptions;

    if (lpFoldWindow=StackGetFoldWindow(&hWindowsStack, hWnd))
    {
      dwOptions=SendMessage(hWnd, AEM_GETOPTIONS, 0, 0);

      if (!bNoPrintCollapsed != !(dwOptions & AECO_NOPRINTCOLLAPSED))
      {
        if (bNoPrintCollapsed)
          SendMessage(hWnd, AEM_SETOPTIONS, AECOOP_OR, AECO_NOPRINTCOLLAPSED);
        else
          SendMessage(hWnd, AEM_SETOPTIONS, AECOOP_XOR, AECO_NOPRINTCOLLAPSED);
        lpFoldWindow->bReturnPrintOption=TRUE;
      }
      else lpFoldWindow->bReturnPrintOption=FALSE;
    }
  }
  else if (uMsg == AEM_ENDPRINTDOC)
  {
    FOLDWINDOW *lpFoldWindow;

    if (lpFoldWindow=StackGetFoldWindow(&hWindowsStack, hWnd))
    {
      if (lpFoldWindow->bReturnPrintOption)
      {
        if (bNoPrintCollapsed)
          SendMessage(hWnd, AEM_SETOPTIONS, AECOOP_XOR, AECO_NOPRINTCOLLAPSED);
        else
          SendMessage(hWnd, AEM_SETOPTIONS, AECOOP_OR, AECO_NOPRINTCOLLAPSED);
      }
    }
  }
  else if (uMsg == WM_PAINT)
  {
    FOLDWINDOW *lpFoldWindow;
    LEVEL *lpCountLevel1=NULL;
    LEVEL *lpCountLevel2=NULL;
    LEVEL *lpCaretLevel=NULL;
    LEVEL *lpLevel=NULL;
    AECHARINDEX ciFirstLine;
    AECHARINDEX ciLastLine;
    AECHARINDEX ciCaret;
    RECT rcDraw;
    HDC hDC;
    HDC hBufferDC;
    HBITMAP hBitmap;
    HBITMAP hBitmapOld=NULL;
    HBRUSH hFirstBkBrush;
    HPEN hSecondBkPen;
    HPEN hNormalFoldPen;
    HPEN hActiveFoldPen;
    HPEN hPenOld;
    int nLastFigure=FRG_NONE;
    int nHPos;
    int nVPos;
    int nCharHeight;
    int i;
    BOOL bCaretLevel=FALSE;
    LRESULT lResult;

    lResult=NewEditProcData->NextProc(hWnd, uMsg, wParam, lParam);

    if (!(lpFoldWindow=StackGetFoldWindow(&hWindowsStack, hWnd)))
      lpFoldWindow=FillFoldsStack(NULL, NULL, hWnd, NULL);

    if (lpFoldWindow)
    {
      if (lpFoldWindow->lpSyntaxFile)
      {
        if (!lpFoldWindow->rcBoard.right)
        {
          //Add rect
          SendMessage(hWnd, AEM_GETRECT, 0, (LPARAM)&rcDraw);
          GetClientRect(hWnd, &lpFoldWindow->rcBoard);
          lpFoldWindow->rcBoard.left=rcDraw.left;
          lpFoldWindow->rcBoard.right=rcDraw.left + BOARD_WIDTH + BOARD_EDGE * 2;
          SetEditRect(hWnd, lpFoldWindow->rcBoard.right - lpFoldWindow->rcBoard.left, 0);
          return lResult;
        }
        SendMessage(hWnd, AEM_GETINDEX, AEGI_FIRSTVISIBLELINE, (LPARAM)&ciFirstLine);
        SendMessage(hWnd, AEM_GETINDEX, AEGI_LASTVISIBLELINE, (LPARAM)&ciLastLine);
        SendMessage(hWnd, AEM_GETINDEX, AEGI_CARETCHAR, (LPARAM)&ciCaret);
        nHPos=lpFoldWindow->rcBoard.left + (lpFoldWindow->rcBoard.right - lpFoldWindow->rcBoard.left) / 2;
        nVPos=SendMessage(hWnd, AEM_VPOSFROMLINE, AECT_CLIENT, ciFirstLine.nLine);
        nCharHeight=SendMessage(hWnd, AEM_GETCHARSIZE, AECS_HEIGHT, 0);

        if (hDC=GetDC(hWnd))
        {
          hBufferDC=CreateCompatibleDC(hDC);
          hBitmap=CreateCompatibleBitmap(hDC, lpFoldWindow->rcBoard.right, lpFoldWindow->rcBoard.bottom);
          hBitmapOld=(HBITMAP)SelectObject(hBufferDC, hBitmap);

          hNormalFoldPen=CreatePen(PS_SOLID, 0, crPanelNormalFold);
          hActiveFoldPen=CreatePen(PS_SOLID, 0, crPanelActiveFold);
          hSecondBkPen=CreatePen(PS_SOLID, 0, crPanelSecondBackground);

          hFirstBkBrush=CreateSolidBrush(crPanelFirstBackground);
          FillRect(hBufferDC, &lpFoldWindow->rcBoard, hFirstBkBrush);

          hPenOld=(HPEN)SelectObject(hBufferDC, hSecondBkPen);
          for (i=0; i < lpFoldWindow->rcBoard.bottom + (lpFoldWindow->rcBoard.right - lpFoldWindow->rcBoard.left); i+=2)
          {
            MoveToEx(hBufferDC, lpFoldWindow->rcBoard.left, i, NULL);
            LineTo(hBufferDC, lpFoldWindow->rcBoard.right, i - (lpFoldWindow->rcBoard.right - lpFoldWindow->rcBoard.left));
          }
          SelectObject(hBufferDC, hNormalFoldPen);

          while (ciFirstLine.nLine <= ciLastLine.nLine)
          {
            if (!StackIsLevelCollapsed(&lpFoldWindow->hLevelStack, &lpCountLevel1, ciFirstLine.nLine))
            {
              if (lpLevel=StackGetLevelByLine(&lpFoldWindow->hLevelStack, &lpCountLevel2, ciFirstLine.nLine, TRUE))
              {
                //Active fold
                if (crPanelNormalFold != crPanelActiveFold)
                {
                  if (!bCaretLevel)
                  {
                     bCaretLevel=TRUE;
                     lpCaretLevel=StackGetLevelByIndex(&lpFoldWindow->hLevelStack, NULL, &ciCaret, TRUE);
                  }
                  if (lpCaretLevel && lpCaretLevel->lpFold->lpMinPoint->ciPoint.nLine <= ciFirstLine.nLine &&
                                      lpCaretLevel->lpFold->lpMaxPoint->ciPoint.nLine >= ciFirstLine.nLine)
                    SelectObject(hBufferDC, hActiveFoldPen);
                  else
                    SelectObject(hBufferDC, hNormalFoldPen);
                }

                //Draw figures
                if (nLastFigure == FRG_FOLDEND)
                {
                  MoveToEx(hBufferDC, nHPos, nVPos - nCharHeight / 2, NULL);
                  LineTo(hBufferDC, nHPos, nVPos);
                }

                if (ciFirstLine.nLine == lpLevel->lpFold->lpMinPoint->ciPoint.nLine)
                {
                  if (nLastFigure)
                  {
                    //Line before square
                    MoveToEx(hBufferDC, nHPos, nVPos, NULL);
                    LineTo(hBufferDC, nHPos, nVPos + nCharHeight / 2 - BOARD_WIDTH / 2);
                  }

                  //Square
                  MoveToEx(hBufferDC, nHPos - BOARD_WIDTH / 2, nVPos + nCharHeight / 2 - BOARD_WIDTH / 2, NULL);
                  LineTo(hBufferDC, nHPos + BOARD_WIDTH / 2, nVPos + nCharHeight / 2 - BOARD_WIDTH / 2);
                  LineTo(hBufferDC, nHPos + BOARD_WIDTH / 2, nVPos + nCharHeight / 2 + BOARD_WIDTH / 2);
                  LineTo(hBufferDC, nHPos - BOARD_WIDTH / 2, nVPos + nCharHeight / 2 + BOARD_WIDTH / 2);
                  LineTo(hBufferDC, nHPos - BOARD_WIDTH / 2, nVPos + nCharHeight / 2 - BOARD_WIDTH / 2);

                  //Line after square
                  MoveToEx(hBufferDC, nHPos, nVPos + nCharHeight / 2 + BOARD_WIDTH / 2, NULL);
                  LineTo(hBufferDC, nHPos, nVPos + nCharHeight);

                  if (lpLevel->lpFold->bCollapse)
                  {
                    //Plus
                    MoveToEx(hBufferDC, nHPos - BOARD_WIDTH / 2 + 2, nVPos + nCharHeight / 2, NULL);
                    LineTo(hBufferDC, nHPos + BOARD_WIDTH / 2 - 1, nVPos + nCharHeight / 2);
                    MoveToEx(hBufferDC, nHPos, nVPos + nCharHeight / 2 - BOARD_WIDTH / 2 + 2, NULL);
                    LineTo(hBufferDC, nHPos, nVPos + nCharHeight / 2 + BOARD_WIDTH / 2 - 1);

                    nLastFigure=FRG_FOLDSTARTPLUS;
                  }
                  else
                  {
                    //Minus
                    MoveToEx(hBufferDC, nHPos - BOARD_WIDTH / 2 + 2, nVPos + nCharHeight / 2, NULL);
                    LineTo(hBufferDC, nHPos + BOARD_WIDTH / 2 - 1, nVPos + nCharHeight / 2);

                    nLastFigure=FRG_FOLDSTARTMINUS;
                  }
                }
                else if (ciFirstLine.nLine == lpLevel->lpFold->lpMaxPoint->ciPoint.nLine)
                {
                  //Line to the right
                  MoveToEx(hBufferDC, nHPos, nVPos, NULL);
                  LineTo(hBufferDC, nHPos, nVPos + nCharHeight / 2);
                  MoveToEx(hBufferDC, nHPos, nVPos + nCharHeight / 2, NULL);
                  LineTo(hBufferDC, nHPos + BOARD_WIDTH / 2, nVPos + nCharHeight / 2);

                  if (lpLevel->nDeep == 0)
                    nLastFigure=FRG_NONE;
                  else
                    nLastFigure=FRG_FOLDEND;
                }
                else
                {
                  if (lpLevel->lpFold->bCollapse)
                  {
                    nVPos-=nCharHeight;
                  }
                  else
                  {
                    //Vertical line
                    MoveToEx(hBufferDC, nHPos, nVPos, NULL);
                    LineTo(hBufferDC, nHPos, nVPos + nCharHeight);
                    nLastFigure=FRG_LINE;
                  }
                }
              }
              else nLastFigure=FRG_NONE;

              nVPos+=nCharHeight;
            }
            ciFirstLine.lpLine=ciFirstLine.lpLine->next;
            ++ciFirstLine.nLine;
          }
          BitBlt(hDC, lpFoldWindow->rcBoard.left, lpFoldWindow->rcBoard.top, lpFoldWindow->rcBoard.right - lpFoldWindow->rcBoard.left, lpFoldWindow->rcBoard.bottom - lpFoldWindow->rcBoard.top, hBufferDC, lpFoldWindow->rcBoard.left, 0, SRCCOPY);

          if (hFirstBkBrush) DeleteObject(hFirstBkBrush);
          if (hPenOld) SelectObject(hBufferDC, hPenOld);
          DeleteObject(hNormalFoldPen);
          DeleteObject(hActiveFoldPen);
          DeleteObject(hSecondBkPen);

          if (hBitmapOld) SelectObject(hBufferDC, hBitmapOld);
          DeleteDC(hBufferDC);
          DeleteObject(hBitmap);
          ReleaseDC(hWnd, hDC);
        }
      }
      else
      {
        if (lpFoldWindow->rcBoard.right)
        {
          //Remove rect
          SetEditRect(lpFoldWindow->hWnd, 0, lpFoldWindow->rcBoard.right - lpFoldWindow->rcBoard.left);
          lpFoldWindow->rcBoard.right=0;
        }
      }
    }
    return lResult;
  }
  else if (uMsg == WM_LBUTTONDOWN ||
           uMsg == WM_LBUTTONDBLCLK)
  {
    FOLDWINDOW *lpFoldWindow;
    LEVEL *lpLevel=NULL;
    LEVEL *lpClick=NULL;
    RECT rcClickable;
    POINT pt;
    int nClickLine;

    pt.x=LOWORD(lParam);
    pt.y=HIWORD(lParam);

    if (lpFoldWindow=StackGetFoldWindow(&hWindowsStack, hWnd))
    {
      if (lpFoldWindow->lpSyntaxFile)
      {
        rcClickable.left=lpFoldWindow->rcBoard.left + BOARD_EDGE;
        rcClickable.top=lpFoldWindow->rcBoard.top;
        rcClickable.right=lpFoldWindow->rcBoard.right - BOARD_EDGE;
        rcClickable.bottom=lpFoldWindow->rcBoard.bottom;

        if (PtInRect(&rcClickable, pt))
        {
          nClickLine=SendMessage(hWnd, AEM_LINEFROMVPOS, AECT_CLIENT, pt.y);

          if (lpLevel=StackGetLevelByLine(&lpFoldWindow->hLevelStack, NULL, nClickLine, TRUE))
          {
            //If folds start at the same line, use deepless
            while (lpLevel)
            {
              if (lpLevel->lpFold->lpMinPoint->ciPoint.nLine == nClickLine &&
                  lpLevel->lpFold->lpMinPoint->ciPoint.nLine != lpLevel->lpFold->lpMaxPoint->ciPoint.nLine)
              {
                lpClick=lpLevel;
              }
              else break;

              lpLevel=lpLevel->lpParent;
            }

            if (lpClick)
            {
              StackSwitchCollapse(&lpFoldWindow->hLevelStack, lpFoldWindow->hWnd, lpClick);
              return 0;
            }
          }
        }
      }
    }
  }

  //Call next procedure
  return NewEditProcData->NextProc(hWnd, uMsg, wParam, lParam);
}

BOOL CALLBACK CurrentCollapseProc(void *lpParameter)
{
  LEVEL *lpLevel;
  AECHARINDEX ciCaret;

  if (lpCurrentFoldWindow)
  {
    SendMessage(lpCurrentFoldWindow->hWnd, AEM_GETINDEX, AEGI_CARETCHAR, (LPARAM)&ciCaret);

    if (lpLevel=StackGetLevelByLine(&lpCurrentFoldWindow->hLevelStack, NULL, ciCaret.nLine, TRUE))
    {
      StackSwitchCollapse(&lpCurrentFoldWindow->hLevelStack, lpCurrentFoldWindow->hWnd, lpLevel);
    }
  }
  return TRUE;
}

BOOL CALLBACK CurrentGoBeginProc(void *lpParameter)
{
  LEVEL *lpLevel;
  AECHARINDEX ciCaret;

  if (lpCurrentFoldWindow)
  {
    SendMessage(lpCurrentFoldWindow->hWnd, AEM_GETINDEX, AEGI_CARETCHAR, (LPARAM)&ciCaret);

    if (lpLevel=StackGetLevelByIndex(&lpCurrentFoldWindow->hLevelStack, NULL, &ciCaret, TRUE))
    {
      FoldInView(lpCurrentFoldWindow, lpLevel, IDM_GOBEGIN);
    }
  }
  return TRUE;
}

BOOL CALLBACK CurrentGoEndProc(void *lpParameter)
{
  LEVEL *lpLevel;
  AECHARINDEX ciCaret;

  if (lpCurrentFoldWindow)
  {
    SendMessage(lpCurrentFoldWindow->hWnd, AEM_GETINDEX, AEGI_CARETCHAR, (LPARAM)&ciCaret);

    if (lpLevel=StackGetLevelByIndex(&lpCurrentFoldWindow->hLevelStack, NULL, &ciCaret, TRUE))
    {
      FoldInView(lpCurrentFoldWindow, lpLevel, IDM_GOEND);
    }
  }
  return TRUE;
}

BOOL CALLBACK CurrentSelectProc(void *lpParameter)
{
  LEVEL *lpLevel;
  AECHARINDEX ciCaret;

  if (lpCurrentFoldWindow)
  {
    SendMessage(lpCurrentFoldWindow->hWnd, AEM_GETINDEX, AEGI_CARETCHAR, (LPARAM)&ciCaret);

    if (lpLevel=StackGetLevelByIndex(&lpCurrentFoldWindow->hLevelStack, NULL, &ciCaret, TRUE))
    {
      FoldSelect(lpCurrentFoldWindow, lpLevel);
    }
  }
  return TRUE;
}

BOOL CALLBACK AllCollapseProc(void *lpParameter)
{
  if (lpCurrentFoldWindow)
  {
    StackSwitchCollapse(&lpCurrentFoldWindow->hLevelStack, lpCurrentFoldWindow->hWnd, NULL);
  }
  return TRUE;
}

BOOL CALLBACK AllNextLevelProc(void *lpParameter)
{
  LEVEL *lpLevel=NULL;
  AECHARINDEX ciCaret;

  if (lpCurrentFoldWindow)
  {
    SendMessage(lpCurrentFoldWindow->hWnd, AEM_GETINDEX, AEGI_CARETCHAR, (LPARAM)&ciCaret);
    if (!StackGetLevelByIndex(&lpCurrentFoldWindow->hLevelStack, &lpLevel, &ciCaret, TRUE))
      if (lpLevel && lpLevel->prev)
        lpLevel=lpLevel->prev;

    if (lpLevel && lpLevel->next)
    {
      lpLevel=lpLevel->next;
      SendMessage(lpCurrentFoldWindow->hWnd, AEM_LOCKSCROLL, SB_BOTH, TRUE);
      SendMessage(lpCurrentFoldWindow->hWnd, AEM_EXSETSEL, (WPARAM)&lpLevel->lpFold->lpMinPoint->ciPoint, (LPARAM)&lpLevel->lpFold->lpMinPoint->ciPoint);
      SendMessage(lpCurrentFoldWindow->hWnd, AEM_LOCKSCROLL, SB_BOTH, FALSE);
      ScrollToPoint(lpCurrentFoldWindow->hWnd, NULL);
    }
  }
  return TRUE;
}

BOOL CALLBACK AllPrevLevelProc(void *lpParameter)
{
  LEVEL *lpLevel=NULL;
  AECHARINDEX ciCaret;

  if (lpCurrentFoldWindow)
  {
    SendMessage(lpCurrentFoldWindow->hWnd, AEM_GETINDEX, AEGI_CARETCHAR, (LPARAM)&ciCaret);
    if (StackGetLevelByIndex(&lpCurrentFoldWindow->hLevelStack, &lpLevel, &ciCaret, TRUE))
      if (lpLevel && lpLevel->next && AEC_IndexCompare(&lpLevel->lpFold->lpMinPoint->ciPoint, &ciCaret))
        lpLevel=lpLevel->next;

    if (lpLevel && lpLevel->prev)
    {
      lpLevel=lpLevel->prev;
      SendMessage(lpCurrentFoldWindow->hWnd, AEM_LOCKSCROLL, SB_BOTH, TRUE);
      SendMessage(lpCurrentFoldWindow->hWnd, AEM_EXSETSEL, (WPARAM)&lpLevel->lpFold->lpMinPoint->ciPoint, (LPARAM)&lpLevel->lpFold->lpMinPoint->ciPoint);
      SendMessage(lpCurrentFoldWindow->hWnd, AEM_LOCKSCROLL, SB_BOTH, FALSE);
      ScrollToPoint(lpCurrentFoldWindow->hWnd, NULL);
    }
  }
  return TRUE;
}

HSYNTAXFILE* StackLoadSyntaxFile(HANDLE hFile, HSTACK *hStack, HSYNTAXFILE *lpSyntaxFile)
{
  EXTENSIONINFO *lpExtElement=NULL;
  FOLDINFO *lpFoldInfo=NULL;
  QUOTEINFO *lpQuoteInfo=NULL;
  wchar_t *wszText;
  wchar_t *wpText;
  wchar_t *wpExtension;
  wchar_t *wpQuoteStart;
  wchar_t *wpQuoteEnd;
  wchar_t *wpFoldStart;
  wchar_t *wpFoldEnd;
  wchar_t *wpDelimiters;
  wchar_t wchEscape;
  DWORD dwFileSize;
  DWORD dwBytesRead;
  DWORD dwUnicodeLen;
  int nExtensionLen;
  int nQuoteStartLen;
  int nQuoteEndLen;
  int nFoldStartLen;
  int nFoldEndLen;
  int nDelimitersLen;
  DWORD dwFlags;
  BOOL bQuoteString;

  if ((dwFileSize=GetFileSize(hFile, NULL)) != INVALID_FILE_SIZE)
  {
    if (dwFileSize >= 2)
    {
      if (wszText=wpText=(wchar_t *)GlobalAlloc(GPTR, dwFileSize + 2))
      {
        if (ReadFile(hFile, wszText, dwFileSize, &dwBytesRead, NULL))
        {
          dwUnicodeLen=dwBytesRead / sizeof(wchar_t);
          wszText[dwUnicodeLen++]='\0';
          if (wszText[0] == 0xFEFF)
          {
            ++wpText;
            --dwUnicodeLen;
          }

          if (!lpSyntaxFile)
            lpSyntaxFile=StackInsertSyntaxFile(hStack);

          if (lpSyntaxFile)
          {
            while (1)
            {
              SectionStart:
              SkipCommentW(&wpText);
              GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL);
              if (!NextLineW(&wpText)) goto FreeText;

              if (!xstrcmpiW(wszBuffer, L"Extensions:"))
              {
                while (1)
                {
                  //Parse line
                  while (1)
                  {
                    lpExtElement=NULL;
                    wpExtension=NULL;

                    //Extension
                    if ((nExtensionLen=GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, &bQuoteString)) || bQuoteString)
                    {
                      if (!bQuoteString && wszBuffer[0] == ';') break;

                      if (wpExtension=(wchar_t *)GlobalAlloc(GPTR, nExtensionLen * sizeof(wchar_t) + 2))
                        memcpy(wpExtension, wszBuffer, nExtensionLen * sizeof(wchar_t) + 2);
                      else
                        break;
                    }
                    else goto SectionStart;

                    //Add to stack
                    if (lpExtElement=StackInsertExtension(&lpSyntaxFile->hExtensionStack, nExtensionLen))
                    {
                      lpExtElement->wpExtension=wpExtension;
                      lpExtElement->nExtensionLen=nExtensionLen;
                    }
                    break;
                  }
                  if (!lpExtElement && wpExtension) GlobalFree((HGLOBAL)wpExtension);
                  if (!NextLineW(&wpText)) goto FreeText;
                }
              }
              else if (!xstrcmpiW(wszBuffer, L"Quotes:"))
              {
                while (1)
                {
                  //Parse line
                  while (1)
                  {
                    lpQuoteInfo=NULL;
                    wpQuoteStart=NULL;
                    wpQuoteEnd=NULL;
                    dwFlags=FIF_MATCHCASE;

                    //Quote flags
                    if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                    {
                      dwFlags=xatoiW(wszBuffer);
                      if (dwFlags == 0 && wszBuffer[0] != '0') break;
                    }
                    else goto SectionStart;

                    //Quote start
                    if ((nQuoteStartLen=GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, &bQuoteString)) || bQuoteString)
                    {
                      if (wpQuoteStart=(wchar_t *)GlobalAlloc(GPTR, nQuoteStartLen * sizeof(wchar_t) + 2))
                        memcpy(wpQuoteStart, wszBuffer, nQuoteStartLen * sizeof(wchar_t) + 2);
                      else
                        break;
                    }
                    else break;

                    //Quote end
                    if ((nQuoteEndLen=GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, &bQuoteString)) || bQuoteString)
                    {
                      if (wpQuoteEnd=(wchar_t *)GlobalAlloc(GPTR, nQuoteEndLen * sizeof(wchar_t) + 2))
                        memcpy(wpQuoteEnd, wszBuffer, nQuoteEndLen * sizeof(wchar_t) + 2);
                      else
                        break;
                    }
                    else break;

                    //Escape
                    if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, &bQuoteString) || bQuoteString)
                      wchEscape=wszBuffer[0];
                    else
                      break;

                    //Add to stack
                    if (lpQuoteInfo=StackInsertQuoteInfo(&lpSyntaxFile->hQuoteStack))
                    {
                      lpQuoteInfo->dwFlags=dwFlags;
                      lpQuoteInfo->wpQuoteStart=wpQuoteStart;
                      lpQuoteInfo->nQuoteStartLen=nQuoteStartLen;
                      lpQuoteInfo->wpQuoteEnd=wpQuoteEnd;
                      lpQuoteInfo->nQuoteEndLen=nQuoteEndLen;
                      lpQuoteInfo->wchEscape=wchEscape;
                    }
                    break;
                  }
                  if (!lpQuoteInfo)
                  {
                    if (wpQuoteStart) GlobalFree((HGLOBAL)wpQuoteStart);
                    if (wpQuoteEnd) GlobalFree((HGLOBAL)wpQuoteEnd);
                  }
                  if (!NextLineW(&wpText)) goto FreeText;
                }
              }
              else if (!xstrcmpiW(wszBuffer, L"Folds:"))
              {
                while (1)
                {
                  //Parse line
                  while (1)
                  {
                    lpFoldInfo=NULL;
                    wpFoldStart=NULL;
                    wpFoldEnd=NULL;
                    wpDelimiters=NULL;
                    dwFlags=FIF_MATCHCASE;

                    //Fold flags
                    if (GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, NULL))
                    {
                      dwFlags=xatoiW(wszBuffer);
                      if (dwFlags == 0 && wszBuffer[0] != '0') break;
                    }
                    else goto SectionStart;

                    //Fold start
                    if ((nFoldStartLen=GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, &bQuoteString)) || bQuoteString)
                    {
                      if (wpFoldStart=(wchar_t *)GlobalAlloc(GPTR, nFoldStartLen * sizeof(wchar_t) + 2))
                        memcpy(wpFoldStart, wszBuffer, nFoldStartLen * sizeof(wchar_t) + 2);
                      else
                        break;
                    }
                    else break;

                    //Fold end
                    if ((nFoldEndLen=GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, &bQuoteString)) || bQuoteString)
                    {
                      if (wpFoldEnd=(wchar_t *)GlobalAlloc(GPTR, nFoldEndLen * sizeof(wchar_t) + 2))
                        memcpy(wpFoldEnd, wszBuffer, nFoldEndLen * sizeof(wchar_t) + 2);
                      else
                        break;
                    }
                    else break;

                    //Delimiters
                    if ((nDelimitersLen=GetWordW(wpText, wszBuffer, MAX_PATH, &wpText, &bQuoteString)) || bQuoteString)
                    {
                      if (wpDelimiters=(wchar_t *)GlobalAlloc(GPTR, nDelimitersLen * sizeof(wchar_t) + 2))
                        memcpy(wpDelimiters, wszBuffer, nDelimitersLen * sizeof(wchar_t) + 2);
                      else
                        break;
                    }
                    else break;

                    //Add to stack
                    if (lpFoldInfo=StackInsertFoldInfo(&lpSyntaxFile->hFoldStack))
                    {
                      lpFoldInfo->dwFlags=dwFlags;
                      lpFoldInfo->wpFoldStart=wpFoldStart;
                      lpFoldInfo->nFoldStartLen=nFoldStartLen;
                      lpFoldInfo->wpFoldEnd=wpFoldEnd;
                      lpFoldInfo->nFoldEndLen=nFoldEndLen;
                      lpFoldInfo->wpDelimiters=wpDelimiters;
                    }
                    break;
                  }
                  if (!lpFoldInfo)
                  {
                    if (wpFoldStart) GlobalFree((HGLOBAL)wpFoldStart);
                    if (wpFoldEnd) GlobalFree((HGLOBAL)wpFoldEnd);
                    if (wpDelimiters) GlobalFree((HGLOBAL)wpDelimiters);
                  }
                  if (!NextLineW(&wpText)) goto FreeText;
                }
              }
              else break;
            }
          }
        }
        FreeText:
        GlobalFree((HGLOBAL)wszText);
      }
    }
  }
  return lpSyntaxFile;
}

HSYNTAXFILE* StackInsertSyntaxFile(HSTACK *hStack)
{
  HSYNTAXFILE *lpElement=NULL;

  StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(HSYNTAXFILE));
  return lpElement;
}

HSYNTAXFILE* StackGetSyntaxFileByExt(HSTACK *hStack, wchar_t *wpExtension, int nExtensionLen)
{
  HSYNTAXFILE *lpElement=(HSYNTAXFILE *)hStack->first;

  while (lpElement)
  {
    if (StackGetExtension(&lpElement->hExtensionStack, wpExtension, nExtensionLen))
      return lpElement;

    lpElement=lpElement->next;
  }
  return NULL;
}

void StackFreeSyntaxFile(HSTACK *hStack)
{
  HSYNTAXFILE *lpElement=(HSYNTAXFILE *)hStack->first;

  while (lpElement)
  {
    StackFreeExtension(&lpElement->hExtensionStack);
    StackFreeQuoteInfo(&lpElement->hQuoteStack);
    StackFreeFoldInfo(&lpElement->hFoldStack);

    lpElement=lpElement->next;
  }
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
}

EXTENSIONINFO* StackInsertExtension(HEXTENSIONSTACK *hStack, int nExtensionLen)
{
  EXTENSIONINFO *lpElement1;
  EXTENSIONINFO *lpElement2=NULL;

  if ((DWORD)nExtensionLen < sizeof(hStack->lpExtensionLens) / sizeof(int))
  {
    if (hStack->lpExtensionLens[nExtensionLen])
    {
      lpElement1=(EXTENSIONINFO *)hStack->lpExtensionLens[nExtensionLen];
    }
    else
    {
      lpElement1=(EXTENSIONINFO *)hStack->first;

      while (lpElement1)
      {
        if (lpElement1->nExtensionLen >= nExtensionLen)
          break;

        lpElement1=lpElement1->next;
      }
    }
    StackInsertBefore((stack **)&hStack->first, (stack **)&hStack->last, (stack *)lpElement1, (stack **)&lpElement2, sizeof(EXTENSIONINFO));

    if (lpElement2)
      hStack->lpExtensionLens[nExtensionLen]=(int)lpElement2;
  }
  return lpElement2;
}

EXTENSIONINFO* StackGetExtension(HEXTENSIONSTACK *hStack, wchar_t *wpExtension, int nExtensionLen)
{
  EXTENSIONINFO *lpElement=NULL;

  if ((DWORD)nExtensionLen < sizeof(hStack->lpExtensionLens) / sizeof(int))
  {
    lpElement=(EXTENSIONINFO *)hStack->lpExtensionLens[nExtensionLen];

    while (lpElement)
    {
      if (lpElement->nExtensionLen != nExtensionLen)
        break;
      if (!xstrcmpiW(lpElement->wpExtension, wpExtension))
        return lpElement;

      lpElement=lpElement->next;
    }
  }
  return NULL;
}

void StackFreeExtension(HEXTENSIONSTACK *hStack)
{
  EXTENSIONINFO *lpElement=(EXTENSIONINFO *)hStack->first;

  while (lpElement)
  {
    GlobalFree((HGLOBAL)lpElement->wpExtension);

    lpElement=lpElement->next;
  }
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
  memset(hStack->lpExtensionLens, 0, MAX_PATH * sizeof(int));
}

MANUALSET* StackInsertManual(HSTACK *hStack)
{
  MANUALSET *lpElement=NULL;

  StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(MANUALSET));

  return lpElement;
}

MANUALSET* StackGetManual(HSTACK *hStack, HWND hWnd)
{
  MANUALSET *lpElement=(MANUALSET *)hStack->first;

  while (lpElement)
  {
    if (lpElement->hWnd == hWnd)
      return lpElement;

    lpElement=lpElement->next;
  }
  return NULL;
}

void StackDeleteManual(HSTACK *hStack, void *ms)
{
  StackDelete((stack **)&hStack->first, (stack **)&hStack->last, (stack *)ms);
}

QUOTEINFO* StackInsertQuoteInfo(HQUOTESTACK *hStack)
{
  QUOTEINFO *lpElement=NULL;

  StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(QUOTEINFO));
  return lpElement;
}

void StackFreeQuoteInfo(HQUOTESTACK *hStack)
{
  QUOTEINFO *lpElement=(QUOTEINFO *)hStack->first;

  while (lpElement)
  {
    GlobalFree((HGLOBAL)lpElement->wpQuoteStart);
    GlobalFree((HGLOBAL)lpElement->wpQuoteEnd);

    lpElement=lpElement->next;
  }
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
}

FOLDINFO* StackInsertFoldInfo(HFOLDSTACK *hStack)
{
  FOLDINFO *lpElement=NULL;

  StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(FOLDINFO));
  return lpElement;
}

void StackFreeFoldInfo(HFOLDSTACK *hStack)
{
  FOLDINFO *lpElement=(FOLDINFO *)hStack->first;

  while (lpElement)
  {
    GlobalFree((HGLOBAL)lpElement->wpFoldStart);
    GlobalFree((HGLOBAL)lpElement->wpFoldEnd);
    GlobalFree((HGLOBAL)lpElement->wpDelimiters);

    lpElement=lpElement->next;
  }
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
}

void StackBeginBoards()
{
  HWND hWndFrame;
  HWND hWndChild;
  RECT rc={0, 0, 1, 1};

  if (!bMDI)
  {
    hWndChild=GetWindow(hMainWnd, GW_CHILD);

    while (hWndChild)
    {
      if (GetWindowLongA(hWndChild, GWL_ID) == ID_EDIT)
        InvalidateRect(hWndChild, &rc, FALSE);

      hWndChild=GetNextWindow(hWndChild, GW_HWNDNEXT);
    }
  }
  else
  {
    hWndFrame=GetWindow(hMdiClient, GW_CHILD);

    while (hWndFrame)
    {
      hWndChild=GetWindow(hWndFrame, GW_CHILD);

      while (hWndChild)
      {
        if (GetWindowLongA(hWndChild, GWL_ID) == ID_EDIT)
          InvalidateRect(hWndChild, &rc, FALSE);

        hWndChild=GetNextWindow(hWndChild, GW_HWNDNEXT);
      }
      hWndFrame=GetNextWindow(hWndFrame, GW_HWNDNEXT);
    }
  }
}

void StackEndBoards(HSTACK *hStack)
{
  FOLDWINDOW *lpElement=(FOLDWINDOW *)hStack->first;

  while (lpElement)
  {
    if (lpElement->rcBoard.right)
    {
      //Remove rect
      SetEditRect(lpElement->hWnd, 0, lpElement->rcBoard.right - lpElement->rcBoard.left);
      lpElement->rcBoard.right=0;
    }
    lpElement=lpElement->next;
  }
}

FOLDWINDOW* StackInsertFoldWindow(HSTACK *hStack)
{
  FOLDWINDOW *lpElement=NULL;

  StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(FOLDWINDOW));

  return lpElement;
}

FOLDWINDOW* StackGetFoldWindow(HSTACK *hStack, HWND hWnd)
{
  FOLDWINDOW *lpElement=(FOLDWINDOW *)hStack->first;

  if (lpCurrentFoldWindow)
  {
    if (lpCurrentFoldWindow->hWnd == hWnd)
      return lpCurrentFoldWindow;
  }

  while (lpElement)
  {
    if (lpElement->hWnd == hWnd)
      return lpElement;

    lpElement=lpElement->next;
  }
  return NULL;
}

void StackDeleteFoldWindow(HSTACK *hStack, FOLDWINDOW *lpElement)
{
  if (lpCurrentFoldWindow == lpElement)
    lpCurrentFoldWindow=NULL;

  StackFreeLevel(&lpElement->hLevelStack, lpElement->hWnd);
  StackDelete((stack **)&hStack->first, (stack **)&hStack->last, (stack *)lpElement);
}

void StackFreeFoldWindows(HSTACK *hStack)
{
  FOLDWINDOW *lpElement=(FOLDWINDOW *)hStack->first;

  while (lpElement)
  {
    StackFreeLevel(&lpElement->hLevelStack, lpElement->hWnd);

    lpElement=lpElement->next;
  }
  lpCurrentFoldWindow=NULL;

  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
}

LEVEL* StackInsertLevel(HSTACK *hLevelStack, AECHARINDEX *ciChar)
{
  LEVEL *lpElement1=(LEVEL *)hLevelStack->last;
  LEVEL *lpElement2=NULL;

  while (lpElement1)
  {
    if (lpElement1->lpFold)
    {
      if (AEC_IndexCompare(ciChar, &lpElement1->lpFold->lpMinPoint->ciPoint) >= 0)
        break;
    }
    else
    {
      if (AEC_IndexCompare(ciChar, &lpElement1->pointMin.ciPoint) >= 0)
        break;
    }

    lpElement1=lpElement1->prev;
  }
  StackInsertAfter((stack **)&hLevelStack->first, (stack **)&hLevelStack->last, (stack *)lpElement1, (stack **)&lpElement2, sizeof(LEVEL));

  if (lpElement2)
  {
    lpElement2->pointMin.nPointOffset=AEPTO_IGNORE;
    lpElement2->pointMax.nPointOffset=AEPTO_IGNORE;
  }
  return lpElement2;
}

LEVEL* StackGetLevelByLine(HSTACK *hLevelStack, LEVEL **lpLevel, int nLine, BOOL bDeepest)
{
  LEVEL *lpCount;
  LEVEL *lpResult=NULL;

  if (!lpLevel || !*lpLevel)
    lpCount=(LEVEL *)hLevelStack->first;
  else
  {
    lpCount=*lpLevel;

    if (lpCount->lpFold->lpMinPoint->ciPoint.nLine > nLine ||
        lpCount->lpFold->lpMaxPoint->ciPoint.nLine < nLine ||
        lpCount->lpFold->lpMinPoint->ciPoint.nLine == lpCount->lpFold->lpMaxPoint->ciPoint.nLine)
    {
      while (lpCount->lpParent)
        lpCount=lpCount->lpParent;
    }
  }

  while (lpCount)
  {
    if (lpCount->lpFold->lpMinPoint->ciPoint.nLine > nLine)
      break;
    if (lpCount->lpFold->lpMaxPoint->ciPoint.nLine >= nLine &&
        lpCount->lpFold->lpMinPoint->ciPoint.nLine != lpCount->lpFold->lpMaxPoint->ciPoint.nLine)
    {
      lpResult=lpCount;
      if (!bDeepest) break;
    }
    lpCount=lpCount->next;
  }

  if (lpLevel)
  {
    if (lpResult)
      *lpLevel=lpResult;
    else
    {
      if (lpCount)
        *lpLevel=lpCount;
      else
        *lpLevel=(LEVEL *)hLevelStack->last;
    }
  }
  return lpResult;
}

LEVEL* StackGetLevelByIndex(HSTACK *hLevelStack, LEVEL **lpLevel, AECHARINDEX *ciChar, BOOL bDeepest)
{
  LEVEL *lpCount;
  LEVEL *lpResult=NULL;

  if (!lpLevel || !*lpLevel)
    lpCount=(LEVEL *)hLevelStack->first;
  else
  {
    lpCount=*lpLevel;

    if (AEC_IndexCompareEx(&lpCount->lpFold->lpMinPoint->ciPoint, ciChar) >= 0 ||
        AEC_IndexCompareEx(&lpCount->lpFold->lpMaxPoint->ciPoint, ciChar) < 0)
    {
      while (lpCount->lpParent)
        lpCount=lpCount->lpParent;
    }
  }

  while (lpCount)
  {
    if (AEC_IndexCompareEx(&lpCount->lpFold->lpMinPoint->ciPoint, ciChar) >= 0)
      break;
    if (AEC_IndexCompareEx(&lpCount->lpFold->lpMaxPoint->ciPoint, ciChar) >= 0)
    {
      lpResult=lpCount;
      if (!bDeepest) break;
    }
    lpCount=lpCount->next;
  }

  if (lpLevel)
  {
    if (lpResult)
      *lpLevel=lpResult;
    else
    {
      if (lpCount)
        *lpLevel=lpCount;
      else
        *lpLevel=(LEVEL *)hLevelStack->last;
    }
  }
  return lpResult;
}

LEVEL* StackFoldAtIndex(HSTACK *hLevelStack, LEVEL *lpLevel, AECHARINDEX *ciChar, DWORD dwFoldStop)
{
  if (!lpLevel)
    lpLevel=(LEVEL *)hLevelStack->first;

  while (lpLevel)
  {
    if (AEC_IndexCompareEx(&lpLevel->lpFold->lpMinPoint->ciPoint, ciChar) > 0)
      break;
    if ((dwFoldStop & IF_FOLDSTART) && !AEC_IndexCompareEx(&lpLevel->lpFold->lpMinPoint->ciPoint, ciChar))
      return lpLevel;
    if ((dwFoldStop & IF_FOLDEND) && !AEC_IndexCompareEx(&lpLevel->lpFold->lpMaxPoint->ciPoint, ciChar))
      return lpLevel;

    lpLevel=lpLevel->next;
  }
  return NULL;
}

BOOL StackIsLevelCollapsed(HSTACK *hLevelStack, LEVEL **lpLevel, int nLine)
{
  LEVEL *lpCount;

  if (!lpLevel || !*lpLevel)
    lpCount=(LEVEL *)hLevelStack->first;
  else
    lpCount=*lpLevel;

  while (lpCount)
  {
    if (lpCount->lpFold->lpMinPoint->ciPoint.nLine > nLine)
      break;

    if (lpCount->lpFold->bCollapse)
    {
      if (lpCount->lpFold->lpMinPoint->ciPoint.nLine + lpCount->lpFold->nHideMinLineOffset <= nLine && nLine <= lpCount->lpFold->lpMaxPoint->ciPoint.nLine + lpCount->lpFold->nHideMaxLineOffset)
      {
        if (lpLevel) *lpLevel=lpCount;
        return TRUE;
      }
      if (lpCount->lpFold->lpMaxPoint->ciPoint.nLine + lpCount->lpFold->nHideMaxLineOffset >= nLine)
        break;
    }
    lpCount=lpCount->next;
  }

  if (lpLevel)
  {
    if (lpCount)
      *lpLevel=lpCount;
    else
      *lpLevel=(LEVEL *)hLevelStack->last;
  }
  return FALSE;
}

void StackSwitchCollapse(HSTACK *hLevelStack, HWND hWnd, LEVEL *lpLevel)
{
  AECHARINDEX ciCaret;
  POINT ptPos;
  POINT *lpPos=&ptPos;

  SendMessage(hWnd, AEM_GETINDEX, AEGI_CARETCHAR, (LPARAM)&ciCaret);

  if (lpLevel)
  {
    if (lpLevel->lpFold->lpMinPoint->ciPoint.nLine + lpLevel->lpFold->nHideMinLineOffset <= ciCaret.nLine &&
        lpLevel->lpFold->lpMaxPoint->ciPoint.nLine + lpLevel->lpFold->nHideMaxLineOffset >= ciCaret.nLine)
    {
      SendMessage(hWnd, AEM_LOCKSCROLL, SB_BOTH, TRUE);
      SendMessage(hWnd, AEM_EXSETSEL, (WPARAM)&lpLevel->lpFold->lpMinPoint->ciPoint, (LPARAM)&lpLevel->lpFold->lpMinPoint->ciPoint);
      SendMessage(hWnd, AEM_LOCKSCROLL, SB_BOTH, FALSE);
      lpPos=NULL;
    }
    else SendMessage(hWnd, AEM_POSFROMCHAR, (WPARAM)lpPos, (LPARAM)&lpLevel->lpFold->lpMinPoint->ciPoint);

    SendMessage(hWnd, AEM_LOCKSCROLL, SB_BOTH, TRUE);
    SendMessage(hWnd, AEM_COLLAPSEFOLD, (WPARAM)lpLevel->lpFold, !lpLevel->lpFold->bCollapse);
    SendMessage(hWnd, AEM_LOCKSCROLL, SB_BOTH, FALSE);
    ScrollToPoint(hWnd, lpPos);
  }
  else
  {
    if (lpLevel=StackGetLevelByLine(hLevelStack, NULL, ciCaret.nLine, TRUE))
    {
      while (lpLevel->lpParent)
        lpLevel=lpLevel->lpParent;

      SendMessage(hWnd, AEM_LOCKSCROLL, SB_BOTH, TRUE);
      SendMessage(hWnd, AEM_EXSETSEL, (WPARAM)&lpLevel->lpFold->lpMinPoint->ciPoint, (LPARAM)&lpLevel->lpFold->lpMinPoint->ciPoint);
      SendMessage(hWnd, AEM_LOCKSCROLL, SB_BOTH, FALSE);
    }
    if (lpLevel=(LEVEL *)hLevelStack->first)
    {
      SendMessage(hWnd, AEM_LOCKSCROLL, SB_BOTH, TRUE);
      SendMessage(hWnd, AEM_COLLAPSEFOLD, (WPARAM)NULL, !lpLevel->lpFold->bCollapse);
      SendMessage(hWnd, AEM_LOCKSCROLL, SB_BOTH, FALSE);
      ScrollToPoint(hWnd, NULL);
    }
  }
}

void StackDeleteLevel(HSTACK *hLevelStack, HWND hWnd, LEVEL *lpLevel)
{
  if (lpLevel->lpFold)
    SendMessage(hWnd, AEM_DELETEFOLD, (WPARAM)lpLevel->lpFold, 0);
  StackDelete((stack **)&hLevelStack->first, (stack **)&hLevelStack->last, (stack *)lpLevel);
}

void StackFreeLevel(HSTACK *hLevelStack, HWND hWnd)
{
  SendMessage(hWnd, AEM_DELETEFOLD, (WPARAM)NULL, 0);
  StackClear((stack **)&hLevelStack->first, (stack **)&hLevelStack->last);
}

FOLDWINDOW* FillFoldsStack(FOLDWINDOW *lpFoldWindow, HSTACK *hLevelStack, HWND hWnd, AECHARRANGE *crRange)
{
  HSYNTAXFILE *lpSyntaxFile=NULL;
  MANUALSET *lpManual;
  AEFINDTEXTW ft;
  AECHARRANGE crScan;
  AECHARINDEX ciCount;
  LEVEL *lpLevel=NULL;
  LEVEL *lpParent=NULL;
  FOLDINFO *lpFoldInfo;
  QUOTEINFO *lpQuoteInfo=NULL;
  EDITINFO ei;
  wchar_t wszExt[MAX_PATH];
  wchar_t *wpExt=NULL;
  int nLevel=0;
  DWORD dwMatchResult;

  if (lpManual=StackGetManual(&hManualStack, hWnd))
  {
    wpExt=lpManual->wszExt;
  }
  else
  {
    if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)hWnd, (LPARAM)&ei))
    {
      if (bOldWindows)
      {
        MultiByteToWideChar(CP_ACP, 0, GetFileExtA((char *)ei.pFile), -1, wszExt, MAX_PATH);
        wpExt=wszExt;
      }
      else wpExt=GetFileExtW((wchar_t *)ei.pFile);
    }
  }
  if (wpExt)
  {
    lpSyntaxFile=StackGetSyntaxFileByExt(&hSyntaxFilesStack, wpExt, lstrlenW(wpExt));
  }

  if (!lpFoldWindow)
  {
    if (lpFoldWindow=StackInsertFoldWindow(&hWindowsStack))
    {
      lpFoldWindow->hWnd=hWnd;
      lpFoldWindow->lpSyntaxFile=lpSyntaxFile;
      hLevelStack=&lpFoldWindow->hLevelStack;
    }
  }
  else
  {
    if (lpFoldWindow->hWnd != hWnd)
      return NULL;
    lpFoldWindow->lpSyntaxFile=lpSyntaxFile;
  }

  if (lpFoldWindow && lpFoldWindow->lpSyntaxFile)
  {
    if (crRange)
    {
      crScan=*crRange;
      //SendMessage(hWnd, AEM_GETINDEX, AEGI_WRAPLINEBEGIN, (LPARAM)&crScan.ciMin);
      //SendMessage(hWnd, AEM_GETINDEX, AEGI_WRAPLINEEND, (LPARAM)&crScan.ciMax);
    }
    else
    {
      SendMessage(hWnd, AEM_GETINDEX, AEGI_FIRSTCHAR, (LPARAM)&crScan.ciMin);
      SendMessage(hWnd, AEM_GETINDEX, AEGI_LASTCHAR, (LPARAM)&crScan.ciMax);
    }
    ciCount=crScan.ciMin;

    while (ciCount.lpLine)
    {
      while (ciCount.nCharInLine < ciCount.lpLine->nLineLen)
      {
        if (!lpParent && AEC_IndexCompare(&ciCount, &crScan.ciMax) >= 0)
          break;

        if (!lpQuoteInfo)
        {
          for (lpQuoteInfo=(QUOTEINFO *)lpFoldWindow->lpSyntaxFile->hQuoteStack.first; lpQuoteInfo; lpQuoteInfo=lpQuoteInfo->next)
          {
            //Quote start
            ft.dwFlags=(lpQuoteInfo->dwFlags & FIF_MATCHCASE)?AEFR_MATCHCASE:0;
            ft.pText=lpQuoteInfo->wpQuoteStart;
            ft.dwTextLen=lpQuoteInfo->nQuoteStartLen;
            ft.nNewLine=AELB_ASIS;
            if (IsMatch(&ft, &ciCount))
            {
              if (!IsEscaped(&ciCount, lpQuoteInfo->wchEscape))
                goto NextChar;
            }
          }

          if (lpFoldInfo=IsFold(lpFoldWindow, lpLevel, &ft, &ciCount, &dwMatchResult))
          {
            if ((lpFoldInfo->dwFlags & FIF_ONLYFOLDSTART) || (dwMatchResult & IF_FOLDEND))
            {
              if (lpLevel)
              {
                lpParent=lpLevel->lpParent;
                lpLevel->pointMax.ciPoint=ft.crFound.ciMin;
                lpLevel->pointMax.nPointOffset=AEPTO_CALC;
                lpLevel->pointMax.nPointLen=!(lpFoldInfo->dwFlags & FIF_ONLYFOLDSTART)?lpLevel->lpFoldInfo->nFoldEndLen:0;
                if (hLevelStack != &lpFoldWindow->hTempStack)
                  lpLevel->lpFold=(AEFOLD *)SendMessage(hWnd, AEM_ADDFOLD, (WPARAM)&lpLevel->pointMin, (LPARAM)&lpLevel->pointMax);

                lpLevel=lpParent;
                nLevel=max(nLevel - 1, 0);
              }
            }
            if ((lpFoldInfo->dwFlags & FIF_ONLYFOLDSTART) || (dwMatchResult & IF_FOLDSTART))
            {
              if (lpLevel=StackInsertLevel(hLevelStack, &ft.crFound.ciMin))
              {
                lpLevel->lpParent=lpParent;
                lpLevel->lpFoldInfo=lpFoldInfo;
                lpLevel->nDeep=nLevel;
                lpLevel->pointMin.ciPoint=ft.crFound.ciMin;
                lpLevel->pointMin.nPointOffset=AEPTO_CALC;
                lpLevel->pointMin.nPointLen=lpLevel->lpFoldInfo->nFoldStartLen;

                if ((lpLevel->lpFoldInfo->dwFlags & FIF_NAMEFROMLEFT) ||
                    (!(lpLevel->lpFoldInfo->dwFlags & FIF_NAMEFROMRIGHT) &&
                       lpLevel->lpFoldInfo->nFoldStartLen == 1))
                {
                  GetFoldName(&lpLevel->pointMin.ciPoint, NULL, lpLevel->wszName, MAX_PATH);
                }
                else xstrcpynW(lpLevel->wszName, lpLevel->pointMin.ciPoint.lpLine->wpLine + lpLevel->pointMin.ciPoint.nCharInLine, MAX_PATH);

                lpParent=lpLevel;
                ++nLevel;
              }
            }
            ciCount=ft.crFound.ciMax;
            continue;
          }
        }
        else
        {
          //Quote end
          if (*lpQuoteInfo->wpQuoteEnd)
          {
            ft.dwFlags=(lpQuoteInfo->dwFlags & FIF_MATCHCASE)?AEFR_MATCHCASE:0;
            ft.pText=lpQuoteInfo->wpQuoteEnd;
            ft.dwTextLen=lpQuoteInfo->nQuoteEndLen;
            ft.nNewLine=AELB_ASIS;
            if (IsMatch(&ft, &ciCount))
            {
              if (!IsEscaped(&ciCount, lpQuoteInfo->wchEscape))
                lpQuoteInfo=NULL;
            }
          }
        }

        NextChar:
        AEC_IndexInc(&ciCount);
      }
      //Force one line quote
      if (ciCount.lpLine->nLineBreak != AELB_WRAP)
        lpQuoteInfo=NULL;
      AEC_NextLine(&ciCount);
    }

    //Close non-closed folds
    SendMessage(hWnd, AEM_GETINDEX, AEGI_LASTCHAR, (LPARAM)&ciCount);

    while (lpLevel)
    {
      lpParent=lpLevel->lpParent;

      if (!lpLevel->lpFold)
      {
        lpLevel->pointMax.ciPoint=ciCount;
        lpLevel->pointMax.nPointOffset=AEPTO_CALC;
        lpLevel->pointMax.nPointLen=0;
        if (hLevelStack != &lpFoldWindow->hTempStack)
          lpLevel->lpFold=(AEFOLD *)SendMessage(hWnd, AEM_ADDFOLD, (WPARAM)&lpLevel->pointMin, (LPARAM)&lpLevel->pointMax);

        nLevel=max(nLevel - 1, 0);
      }
      lpLevel=lpParent;
    }
  }
  return lpFoldWindow;
}

void AddFoldsToStack(HSTACK *hLevelStack, HWND hWnd)
{
  LEVEL *lpLevel=(LEVEL *)hLevelStack->first;

  while (lpLevel)
  {
    if (!lpLevel->lpFold)
      lpLevel->lpFold=(AEFOLD *)SendMessage(hWnd, AEM_ADDFOLD, (WPARAM)&lpLevel->pointMin, (LPARAM)&lpLevel->pointMax);

    lpLevel=lpLevel->next;
  }
}

BOOL GetFoldName(const AECHARINDEX *ciInput, AECHARRANGE *crRange, wchar_t *wszName, int nMaxName)
{
  AECHARRANGE cr;
  int nNameLen;

  for (cr.ciMax=*ciInput; AEC_PrevChar(&cr.ciMax); )
  {
    if (cr.ciMax.nCharInLine < cr.ciMax.lpLine->nLineLen)
    {
      if (cr.ciMax.lpLine->wpLine[cr.ciMax.nCharInLine] != ' ' &&
          cr.ciMax.lpLine->wpLine[cr.ciMax.nCharInLine] != '\t')
      {
        cr.ciMin.nCharInLine=0;
        cr.ciMin.lpLine=cr.ciMax.lpLine;
        cr.ciMin.nLine=cr.ciMax.nLine;

        while (cr.ciMin.lpLine->wpLine[cr.ciMin.nCharInLine] == ' ' ||
               cr.ciMin.lpLine->wpLine[cr.ciMin.nCharInLine] == '\t')
        {
          AEC_NextChar(&cr.ciMin);
        }
        nNameLen=min(cr.ciMax.nCharInLine - cr.ciMin.nCharInLine + 1, nMaxName - 2);
        if (wszName) xstrcpynW(wszName, cr.ciMin.lpLine->wpLine + cr.ciMin.nCharInLine, nNameLen + 1);
        if (crRange) *crRange=cr;
        return nNameLen;
      }
    }
  }
  return 0;
}

BOOL FoldInView(FOLDWINDOW *lpFoldWindow, LEVEL *lpLevel, int nMenuAction)
{
  AEINDEXOFFSET io;
  AECHARRANGE cr;
  POINT pt;
  BOOL bResult=TRUE;

  if (nMenuAction == IDM_GOEND)
  {
    cr.ciMin=lpLevel->lpFold->lpMaxPoint->ciPoint;

    io.ciCharIn=&cr.ciMin;
    io.ciCharOut=&cr.ciMin;
    io.nOffset=lpLevel->lpFold->lpMaxPoint->nPointLen;
    io.nNewLine=AELB_R;
    SendMessage(lpFoldWindow->hWnd, AEM_INDEXOFFSET, 0, (LPARAM)&io);
  }
  else
  {
    if ((lpLevel->lpFoldInfo->dwFlags & FIF_NAMEFROMLEFT) ||
        (!(lpLevel->lpFoldInfo->dwFlags & FIF_NAMEFROMRIGHT) &&
           lpLevel->lpFoldInfo->nFoldStartLen == 1))
    {
      bResult=GetFoldName(&lpLevel->lpFold->lpMinPoint->ciPoint, &cr, NULL, 0);
    }
    else cr.ciMin=lpLevel->lpFold->lpMinPoint->ciPoint;
  }

  if (bResult)
  {
    if (nMenuAction == IDM_INVIEW)
    {
      SendMessage(lpFoldWindow->hWnd, AEM_POSFROMCHAR, (WPARAM)&pt, (LPARAM)&cr.ciMin);
      SendMessage(lpFoldWindow->hWnd, AEM_CONVERTPOINT, AECPT_CLIENTTOGLOBAL, (LPARAM)&pt);
      SendMessage(lpFoldWindow->hWnd, AEM_SETSCROLLPOS, 0, (LPARAM)&pt);
    }
    else
    {
      SendMessage(lpFoldWindow->hWnd, AEM_LOCKSCROLL, SB_BOTH, TRUE);
      lpFoldWindow->bLockSel=TRUE;
      SendMessage(lpFoldWindow->hWnd, AEM_EXSETSEL, (WPARAM)&cr.ciMin, (LPARAM)&cr.ciMin);
      lpFoldWindow->bLockSel=FALSE;
      SendMessage(lpFoldWindow->hWnd, AEM_LOCKSCROLL, SB_BOTH, FALSE);
      ScrollToPoint(lpFoldWindow->hWnd, NULL);
    }
  }
  return bResult;
}


BOOL FoldSelect(FOLDWINDOW *lpFoldWindow, LEVEL *lpLevel)
{
  AEINDEXOFFSET io;
  AECHARRANGE crRange;

  if ((lpLevel->lpFoldInfo->dwFlags & FIF_NAMEFROMLEFT) ||
      (!(lpLevel->lpFoldInfo->dwFlags & FIF_NAMEFROMRIGHT) &&
         lpLevel->lpFoldInfo->nFoldStartLen == 1))
  {
    if (!GetFoldName(&lpLevel->lpFold->lpMinPoint->ciPoint, &crRange, NULL, 0))
      crRange.ciMin=lpLevel->lpFold->lpMinPoint->ciPoint;
    crRange.ciMax=lpLevel->lpFold->lpMaxPoint->ciPoint;
  }
  else
  {
    crRange.ciMin=lpLevel->lpFold->lpMinPoint->ciPoint;
    crRange.ciMax=lpLevel->lpFold->lpMaxPoint->ciPoint;
  }
  io.ciCharIn=&crRange.ciMax;
  io.ciCharOut=&crRange.ciMax;
  io.nOffset=lpLevel->lpFold->lpMaxPoint->nPointLen;
  io.nNewLine=AELB_R;
  SendMessage(lpFoldWindow->hWnd, AEM_INDEXOFFSET, 0, (LPARAM)&io);

  //Set selection
  SendMessage(lpFoldWindow->hWnd, AEM_LOCKSCROLL, SB_BOTH, TRUE);
  lpFoldWindow->bLockSel=TRUE;
  SendMessage(lpFoldWindow->hWnd, AEM_EXSETSEL, (WPARAM)&crRange.ciMax, (LPARAM)&crRange.ciMin);
  lpFoldWindow->bLockSel=FALSE;
  SendMessage(lpFoldWindow->hWnd, AEM_LOCKSCROLL, SB_BOTH, FALSE);
  ScrollToPoint(lpFoldWindow->hWnd, NULL);
  return TRUE;
}

void FillTreeViewA(HWND hWndTreeView, FOLDWINDOW *lpFoldWindow)
{
  LEVEL *lpElement=(LEVEL *)lpFoldWindow->hLevelStack.first;
  TVINSERTSTRUCTA tvis;
  TVITEMA tvi;
  char szName[MAX_PATH];

  SendMessage(hWndTreeView, WM_SETREDRAW, FALSE, 0);
  SendMessage(hWndTreeView, TVM_DELETEITEM, 0, (LPARAM)NULL);

  while (lpElement)
  {
    WideCharToMultiByte(CP_ACP, 0, lpElement->wszName, -1, szName, MAX_PATH, NULL, NULL);
    tvi.mask=TVIF_TEXT|TVIF_PARAM;
    tvi.pszText=szName;
    tvi.cchTextMax=MAX_PATH;
    tvi.lParam=(LPARAM)lpElement;

    tvis.item=tvi;
    tvis.hInsertAfter=TVI_LAST;
    tvis.hParent=lpElement->lpParent?lpElement->lpParent->hItem:NULL;
    lpElement->hItem=(HTREEITEM)SendMessage(hWndTreeView, TVM_INSERTITEMA, 0, (LPARAM)&tvis);

    lpElement=lpElement->next;
  }
  SendMessage(hWndTreeView, WM_SETREDRAW, TRUE, 0);
  InvalidateRect(hWndTreeView, NULL, TRUE);
}

void FillTreeViewW(HWND hWndTreeView, FOLDWINDOW *lpFoldWindow)
{
  LEVEL *lpElement=(LEVEL *)lpFoldWindow->hLevelStack.first;
  TVINSERTSTRUCTW tvis;
  TVITEMW tvi;

  SendMessage(hWndTreeView, WM_SETREDRAW, FALSE, 0);
  SendMessage(hWndTreeView, TVM_DELETEITEM, 0, (LPARAM)NULL);

  while (lpElement)
  {
    tvi.mask=TVIF_TEXT|TVIF_PARAM;
    tvi.pszText=lpElement->wszName;
    tvi.cchTextMax=MAX_PATH;
    tvi.lParam=(LPARAM)lpElement;

    tvis.item=tvi;
    tvis.hInsertAfter=TVI_LAST;
    tvis.hParent=lpElement->lpParent?lpElement->lpParent->hItem:NULL;
    lpElement->hItem=(HTREEITEM)SendMessage(hWndTreeView, TVM_INSERTITEMW, 0, (LPARAM)&tvis);

    lpElement=lpElement->next;
  }
  SendMessage(hWndTreeView, WM_SETREDRAW, TRUE, 0);
  InvalidateRect(hWndTreeView, NULL, TRUE);
}

void FilterTreeViewA(HWND hWndTreeView, FOLDWINDOW *lpFoldWindow, char *pFindIt)
{
  LEVEL *lpElement=(LEVEL *)lpFoldWindow->hLevelStack.first;
  TVINSERTSTRUCTA tvis;
  TVITEMA tvi;
  char szName[MAX_PATH];
  wchar_t wszFindIt[MAX_PATH];

  SendMessage(hWndTreeView, WM_SETREDRAW, FALSE, 0);
  SendMessage(hWndTreeView, TVM_DELETEITEM, 0, (LPARAM)NULL);
  MultiByteToWideChar(CP_ACP, 0, pFindIt, -1, wszFindIt, MAX_PATH);

  while (lpElement)
  {
    if (xstrstrW(lpElement->wszName, (DWORD)-1, wszFindIt, FALSE, NULL, NULL))
    {
      WideCharToMultiByte(CP_ACP, 0, lpElement->wszName, -1, szName, MAX_PATH, NULL, NULL);
      tvi.mask=TVIF_TEXT|TVIF_PARAM;
      tvi.pszText=szName;
      tvi.cchTextMax=MAX_PATH;
      tvi.lParam=(LPARAM)lpElement;

      tvis.item=tvi;
      tvis.hInsertAfter=TVI_LAST;
      tvis.hParent=NULL;
      SendMessage(hWndTreeView, TVM_INSERTITEMA, 0, (LPARAM)&tvis);
    }
    lpElement=lpElement->next;
  }
  SendMessage(hWndTreeView, WM_SETREDRAW, TRUE, 0);
  InvalidateRect(hWndTreeView, NULL, TRUE);
}

void FilterTreeViewW(HWND hWndTreeView, FOLDWINDOW *lpFoldWindow, wchar_t *wpFindIt)
{
  LEVEL *lpElement=(LEVEL *)lpFoldWindow->hLevelStack.first;
  TVINSERTSTRUCTW tvis;
  TVITEMW tvi;

  SendMessage(hWndTreeView, WM_SETREDRAW, FALSE, 0);
  SendMessage(hWndTreeView, TVM_DELETEITEM, 0, (LPARAM)NULL);

  while (lpElement)
  {
    if (xstrstrW(lpElement->wszName, (DWORD)-1, wpFindIt, FALSE, NULL, NULL))
    {
      tvi.mask=TVIF_TEXT|TVIF_PARAM;
      tvi.pszText=lpElement->wszName;
      tvi.cchTextMax=MAX_PATH;
      tvi.lParam=(LPARAM)lpElement;

      tvis.item=tvi;
      tvis.hInsertAfter=TVI_LAST;
      tvis.hParent=NULL;
      SendMessage(hWndTreeView, TVM_INSERTITEMW, 0, (LPARAM)&tvis);
    }
    lpElement=lpElement->next;
  }
  SendMessage(hWndTreeView, WM_SETREDRAW, TRUE, 0);
  InvalidateRect(hWndTreeView, NULL, TRUE);
}

void ExpandTreeView(HWND hWndTreeView, FOLDWINDOW *lpFoldWindow, BOOL bExpand)
{
  LEVEL *lpElement=(LEVEL *)lpFoldWindow->hLevelStack.first;

  SendMessage(hWndTreeView, WM_SETREDRAW, FALSE, 0);

  while (lpElement)
  {
    if (bExpand)
      SendMessage(hWndTreeView, TVM_EXPAND, TVE_EXPAND, (LPARAM)lpElement->hItem);
    else
      SendMessage(hWndTreeView, TVM_EXPAND, TVE_COLLAPSE, (LPARAM)lpElement->hItem);

    lpElement=lpElement->next;
  }
  SendMessage(hWndTreeView, WM_SETREDRAW, TRUE, 0);
  InvalidateRect(hWndTreeView, NULL, TRUE);
}

LEVEL* GetCurLevelTreeView(HWND hWndTreeView)
{
  TVITEMA tvi;

  tvi.mask=TVIF_PARAM;
  tvi.hItem=(HTREEITEM)SendMessage(hWndTreeView, TVM_GETNEXTITEM, TVGN_CARET, (LPARAM)NULL);

  if (SendMessage(hWndTreeView, TVM_GETITEMA, 0, (LPARAM)&tvi))
  {
    return (LEVEL *)tvi.lParam;
  }
  return NULL;
}

BOOL SetCaretTreeView(HWND hWndTreeView, FOLDWINDOW *lpFoldWindow)
{
  LEVEL *lpLevel;
  AECHARINDEX ciCaret;

  if (bFollowCaret)
  {
    SendMessage(lpFoldWindow->hWnd, AEM_GETINDEX, AEGI_CARETCHAR, (LPARAM)&ciCaret);

    if (lpLevel=StackGetLevelByIndex(&lpFoldWindow->hLevelStack, NULL, &ciCaret, TRUE))
    {
      SendMessage(hWndTreeView, TVM_SELECTITEM, TVGN_CARET, (LPARAM)lpLevel->hItem);
      return TRUE;
    }
  }
  return FALSE;
}

FOLDWINDOW* SetActiveEdit(HWND hWndEdit, HWND hWndTreeView, DWORD dwFlags)
{
  FOLDWINDOW *lpFoldWindow;

  if (!hWndEdit) return NULL;

  if (lpFoldWindow=StackGetFoldWindow(&hWindowsStack, hWndEdit))
  {
    if (dwFlags & SAE_RESETFOLD)
    {
      StackFreeLevel(&lpFoldWindow->hLevelStack, lpFoldWindow->hWnd);
      FillFoldsStack(lpFoldWindow, &lpFoldWindow->hLevelStack, hWndEdit, NULL);
    }
  }
  else lpFoldWindow=FillFoldsStack(NULL, NULL, hWndEdit, NULL);

  if (bShowDock)
  {
    if (lpFoldWindow && hWndTreeView)
    {
      if (dwFlags & SAE_RESETLIST)
      {
        if (bOldWindows)
          FillTreeViewA(hWndTreeView, lpFoldWindow);
        else
          FillTreeViewW(hWndTreeView, lpFoldWindow);
        SetCaretTreeView(hWndTreeView, lpFoldWindow);
      }
    }
  }
  return lpFoldWindow;
}

DWORD IsMatch(AEFINDTEXTW *ft, const AECHARINDEX *ciChar)
{
  AECHARINDEX ciCount=*ciChar;
  DWORD dwCount=0;

  while (1)
  {
    for (; ciCount.nCharInLine < ciCount.lpLine->nLineLen; ++ciCount.nCharInLine)
    {
      if (((ft->dwFlags & AEFR_MATCHCASE) && ciCount.lpLine->wpLine[ciCount.nCharInLine] == ft->pText[dwCount]) ||
          (!(ft->dwFlags & AEFR_MATCHCASE) && WideCharUpper(ciCount.lpLine->wpLine[ciCount.nCharInLine]) == WideCharUpper(ft->pText[dwCount])))
      {
        if (ft->dwTextLen == (DWORD)-1)
        {
          if (!ft->pText[++dwCount])
          {
            ++ciCount.nCharInLine;
            goto Founded;
          }
        }
        else if (++dwCount >= ft->dwTextLen)
        {
          ++ciCount.nCharInLine;
          goto Founded;
        }
      }
      else return FALSE;
    }
    if (ciCount.lpLine->nLineBreak == AELB_EOF)
      return 0;
    if (ciCount.lpLine->nLineBreak != AELB_WRAP)
      return 0;

    if (ciCount.lpLine->next)
      AEC_NextLine(&ciCount);
    else
      return 0;

    if (ft->dwTextLen == (DWORD)-1)
    {
      if (!ft->pText[dwCount])
        goto Founded;
    }
    else
    {
      if (dwCount >= ft->dwTextLen)
        goto Founded;
    }
  }

  Founded:
  ft->crFound.ciMin=*ciChar;
  ft->crFound.ciMax=ciCount;
  return dwCount;
}

BOOL IsEscaped(const AECHARINDEX *ciChar, wchar_t wchEscape)
{
  AECHARINDEX ciCount=*ciChar;
  int nEscapeCount=0;

  while (AEC_PrevCharInLine(&ciCount))
  {
    if (*(ciCount.lpLine->wpLine + ciCount.nCharInLine) == wchEscape)
      ++nEscapeCount;
    else
      break;
  }
  if (nEscapeCount % 2)
    return TRUE;
  return FALSE;
}

FOLDINFO* IsFold(FOLDWINDOW *lpFoldWindow, LEVEL *lpLevel, AEFINDTEXTW *ft, AECHARINDEX *ciChar, DWORD *dwFoldStop)
{
  FOLDINFO *lpFoldInfo=(FOLDINFO *)lpFoldWindow->lpSyntaxFile->hFoldStack.first;

  ft->nNewLine=AELB_ASIS;

  while (lpFoldInfo)
  {
    ft->dwFlags=(lpFoldInfo->dwFlags & FIF_MATCHCASE)?AEFR_MATCHCASE:0;

    //Fold start
    ft->pText=lpFoldInfo->wpFoldStart;
    ft->dwTextLen=lpFoldInfo->nFoldStartLen;

    if (IsMatch(ft, ciChar))
    {
      if (CheckFlags(lpFoldInfo, &ft->crFound, IF_FOLDSTART))
      {
        if (dwFoldStop) *dwFoldStop=IF_FOLDSTART;
        return lpFoldInfo;
      }
    }

    //Fold end
    if (lpLevel && lpFoldInfo == lpLevel->lpFoldInfo)
    {
      if (!(lpFoldInfo->dwFlags & FIF_ONLYFOLDSTART))
      {
        ft->pText=lpFoldInfo->wpFoldEnd;
        ft->dwTextLen=lpFoldInfo->nFoldEndLen;

        if (IsMatch(ft, ciChar))
        {
          if (CheckFlags(lpFoldInfo, &ft->crFound, IF_FOLDEND))
          {
            if (dwFoldStop) *dwFoldStop=IF_FOLDEND;
            return lpFoldInfo;
          }
        }
      }
    }

    lpFoldInfo=lpFoldInfo->next;
  }
  return NULL;
}

FOLDINFO* FindFold(FOLDWINDOW *lpFoldWindow, LEVEL *lpStartScan, AECHARRANGE *crSearchRange)
{
  LEVEL *lpLevel;
  FOLDINFO *lpFoldInfo;
  QUOTEINFO *lpQuoteInfo=NULL;
  AEFINDTEXTW ft;
  AECHARINDEX ciCount=crSearchRange->ciMin;

  while (ciCount.lpLine)
  {
    while (ciCount.nCharInLine < ciCount.lpLine->nLineLen)
    {
      if (AEC_IndexCompare(&ciCount, &crSearchRange->ciMax) >= 0)
        return NULL;

      if (!lpQuoteInfo)
      {
        for (lpQuoteInfo=(QUOTEINFO *)lpFoldWindow->lpSyntaxFile->hQuoteStack.first; lpQuoteInfo; lpQuoteInfo=lpQuoteInfo->next)
        {
          //Quote start
          ft.dwFlags=(lpQuoteInfo->dwFlags & FIF_MATCHCASE)?AEFR_MATCHCASE:0;
          ft.pText=lpQuoteInfo->wpQuoteStart;
          ft.dwTextLen=lpQuoteInfo->nQuoteStartLen;
          ft.nNewLine=AELB_ASIS;
          if (IsMatch(&ft, &ciCount))
          {
            if (!IsEscaped(&ciCount, lpQuoteInfo->wchEscape))
              goto NextChar;
          }
        }

        for (lpFoldInfo=(FOLDINFO *)lpFoldWindow->lpSyntaxFile->hFoldStack.first; lpFoldInfo; lpFoldInfo=lpFoldInfo->next)
        {
          ft.dwFlags=(lpFoldInfo->dwFlags & FIF_MATCHCASE)?AEFR_MATCHCASE:0;
          ft.nNewLine=AELB_ASIS;

          //Fold start
          ft.pText=lpFoldInfo->wpFoldStart;
          ft.dwTextLen=lpFoldInfo->nFoldStartLen;
          if (IsMatch(&ft, &ciCount))
          {
            if (!CheckFlags(lpFoldInfo, &ft.crFound, IF_FOLDSTART) != !StackFoldAtIndex(&lpFoldWindow->hLevelStack, lpStartScan, &ft.crFound.ciMin, IF_FOLDSTART))
              return lpFoldInfo;
          }

          //Fold end
          ft.pText=lpFoldInfo->wpFoldEnd;
          ft.dwTextLen=lpFoldInfo->nFoldEndLen;
          if (IsMatch(&ft, &ciCount))
          {
            if (!CheckFlags(lpFoldInfo, &ft.crFound, IF_FOLDEND) != !StackFoldAtIndex(&lpFoldWindow->hLevelStack, lpStartScan, &ft.crFound.ciMin, IF_FOLDEND))
              return lpFoldInfo;
          }
        }
      }
      else
      {
        //Quote end
        if (*lpQuoteInfo->wpQuoteEnd)
        {
          ft.dwFlags=(lpQuoteInfo->dwFlags & FIF_MATCHCASE)?AEFR_MATCHCASE:0;
          ft.pText=lpQuoteInfo->wpQuoteEnd;
          ft.dwTextLen=lpQuoteInfo->nQuoteEndLen;
          ft.nNewLine=AELB_ASIS;
          if (IsMatch(&ft, &ciCount))
          {
            if (!IsEscaped(&ciCount, lpQuoteInfo->wchEscape))
              lpQuoteInfo=NULL;
          }
        }
        if (lpLevel=StackFoldAtIndex(&lpFoldWindow->hLevelStack, lpStartScan, &ciCount, IF_FOLDSTART|IF_FOLDEND))
          return lpLevel->lpFoldInfo;
      }

      NextChar:
      AEC_IndexInc(&ciCount);
    }
    //Force one line quote
    if (ciCount.lpLine->nLineBreak != AELB_WRAP)
      lpQuoteInfo=NULL;
    AEC_NextLine(&ciCount);
  }
  return NULL;
}

BOOL CheckFlags(FOLDINFO *lpFoldInfo, AECHARRANGE *crFound, DWORD dwFoldStop)
{
  if (dwFoldStop & IF_FOLDSTART)
  {
    if (lpFoldInfo->dwFlags & FIF_FOLDSTARTATLINESTART)
    {
      if (!AEC_IsSpacesFromLeft(&crFound->ciMin))
        return FALSE;
    }
    if (lpFoldInfo->dwFlags & FIF_FOLDSTARTATLINEEND)
    {
      if (!AEC_IsSpacesFromRight(&crFound->ciMax))
        return FALSE;
    }
    if (*lpFoldInfo->wpDelimiters)
    {
      AECHARINDEX ciTmp=crFound->ciMin;

      if (AEC_PrevChar(&ciTmp))
      {
        if (!IsDelimiter(lpFoldInfo->wpDelimiters, AEC_CharAtIndex(&ciTmp)))
          return FALSE;
      }
      if (!IsDelimiter(lpFoldInfo->wpDelimiters, AEC_CharAtIndex(&crFound->ciMax)))
        return FALSE;
    }
  }
  if (dwFoldStop & IF_FOLDEND)
  {
    if (lpFoldInfo->dwFlags & FIF_FOLDENDATLINESTART)
    {
      if (!AEC_IsSpacesFromLeft(&crFound->ciMin))
        return FALSE;
    }
    if (lpFoldInfo->dwFlags & FIF_FOLDENDATLINEEND)
    {
      if (!AEC_IsSpacesFromRight(&crFound->ciMax))
        return FALSE;
    }
    if (*lpFoldInfo->wpDelimiters)
    {
      AECHARINDEX ciTmp=crFound->ciMin;

      if (AEC_PrevChar(&ciTmp))
      {
        if (!IsDelimiter(lpFoldInfo->wpDelimiters, AEC_CharAtIndex(&ciTmp)))
          return FALSE;
      }
      if (!IsDelimiter(lpFoldInfo->wpDelimiters, AEC_CharAtIndex(&crFound->ciMax)))
        return FALSE;
    }
  }
  return TRUE;
}

BOOL AEC_IsSpacesFromLeft(const AECHARINDEX *ciChar)
{
  AECHARINDEX ciCount=*ciChar;
  wchar_t wchChar;

  if (AEC_IsFirstCharInLine(&ciCount))
    return TRUE;

  while (AEC_PrevCharInLine(&ciCount))
  {
    wchChar=*(ciCount.lpLine->wpLine + ciCount.nCharInLine);
    if (wchChar != L' ' && wchChar != L'\t')
      return FALSE;
  }
  return TRUE;
}

BOOL AEC_IsSpacesFromRight(const AECHARINDEX *ciChar)
{
  AECHARINDEX ciCount=*ciChar;
  wchar_t wchChar;

  if (AEC_IsLastCharInLine(&ciCount))
    return TRUE;

  do
  {
    wchChar=*(ciCount.lpLine->wpLine + ciCount.nCharInLine);
    if (wchChar != L' ' && wchChar != L'\t')
      return FALSE;
  }
  while (AEC_NextCharInLine(&ciCount));

  return TRUE;
}

void SetEditRect(HWND hWnd, int nNewWidth, int nOldWidth)
{
  RECT rcClient;
  RECT rcDraw;
  int nBorderX;
  int nBorderY;

  GetClientRect(hWnd, &rcClient);
  SendMessage(hWnd, EM_GETRECT, 0, (LPARAM)&rcDraw);

  rcClient.left=nNewWidth + (rcDraw.left - nOldWidth);
  rcClient.right-=(rcClient.right - rcDraw.right);

  //if (bOldRichEdit)
  //{
  //  rcClient.left+=3;
  //  rcClient.top+=4;
  //  rcClient.right-=3;
  //  rcClient.bottom-=4;
  //}
  //else
  {
    if ((GetWindowLong(hWnd, GWL_EXSTYLE) & WS_EX_CLIENTEDGE) ||
        (GetWindowLong(hWnd, GWL_STYLE) & ES_SUNKEN))
    {
      nBorderX=GetSystemMetrics(SM_CXBORDER);
      nBorderY=GetSystemMetrics(SM_CYBORDER);

      rcClient.left+=nBorderX;
      rcClient.top+=nBorderY;
      rcClient.right-=nBorderX;
      rcClient.bottom-=nBorderY;
    }
  }

  SendMessage(hWnd, EM_SETRECT, 0, (LPARAM)&rcClient);
}

HWND GetCurEdit()
{
 EDITINFO ei;

 if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
   return ei.hWndEdit;
 return NULL;
}

DWORD ScrollToPoint(HWND hWnd, POINT *ptPos)
{
  AESCROLLTOPOINT stp;
  DWORD dwScrollResult;

  //Test scroll to caret
  stp.dwFlags=AESC_TEST|(ptPos?AESC_POINTCLIENT:AESC_POINTCARET)|AESC_OFFSETCHARX|AESC_OFFSETCHARY;
  stp.ptPos.x=ptPos?ptPos->x:0;
  stp.ptPos.y=ptPos?ptPos->y:0;
  stp.nOffsetX=1;
  stp.nOffsetY=0;
  dwScrollResult=SendMessage(hWnd, AEM_SCROLLTOPOINT, 0, (LPARAM)&stp);

  //Scroll to caret
  stp.dwFlags=(ptPos?AESC_POINTCLIENT:AESC_POINTCARET);
  stp.nOffsetX=3;
  stp.nOffsetY=2;
  if (dwScrollResult & AECSE_SCROLLEDX)
    stp.dwFlags|=AESC_OFFSETRECTDIVX;
  if (dwScrollResult & AECSE_SCROLLEDY)
    stp.dwFlags|=AESC_OFFSETRECTDIVY;
  return SendMessage(hWnd, AEM_SCROLLTOPOINT, 0, (LPARAM)&stp);
}

int GetWordA(char *pText, char *szWord, int nWordLenMax, char **pNextWord, BOOL *bQuote)
{
  char *pCount;
  char chStopChar;
  int nWordLen;

  while (*pText == ' ' || *pText == '\t') ++pText;

  if (*pText == '\"' || *pText == '\'' || *pText == '`')
  {
    //Parse: "param" or 'param' or `param`
    chStopChar=*pText;
    pCount=++pText;
    while (*pCount != chStopChar && *pCount != '\r' && *pCount != '\0')
      ++pCount;

    if (pNextWord)
    {
      *pNextWord=pCount;
      if (*pCount == chStopChar)
        ++*pNextWord;
    }
    if (bQuote) *bQuote=TRUE;
  }
  else
  {
    //Parse: param1 param2 param3
    pCount=pText;

    while (*pCount != ' ' && *pCount != '\t' && *pCount != '\r' && *pCount != '\0')
      ++pCount;

    if (pNextWord)
    {
      *pNextWord=pCount;
    }
    if (bQuote) *bQuote=FALSE;
  }
  nWordLen=min(pCount - pText, nWordLenMax - 1);
  memcpy(szWord, pText, nWordLen);
  szWord[nWordLen]='\0';

  return nWordLen;
}

int GetWordW(wchar_t *wpText, wchar_t *wszWord, int nWordLenMax, wchar_t **wpNextWord, BOOL *bQuote)
{
  wchar_t *wpCount;
  wchar_t wchStopChar;
  int nWordLen;

  while (*wpText == ' ' || *wpText == '\t') ++wpText;

  if (*wpText == '\"' || *wpText == '\'' || *wpText == '`')
  {
    //Parse: "param" or 'param' or `param`
    wchStopChar=*wpText;
    wpCount=++wpText;
    while (*wpCount != wchStopChar && *wpCount != '\r' && *wpCount != '\0')
      ++wpCount;

    if (wpNextWord)
    {
      *wpNextWord=wpCount;
      if (*wpCount == wchStopChar)
        ++*wpNextWord;
    }
    if (bQuote) *bQuote=TRUE;
  }
  else
  {
    //Parse: param1 param2 param3
    wpCount=wpText;

    while (*wpCount != ' ' && *wpCount != '\t' && *wpCount != '\r' && *wpCount != '\0')
      ++wpCount;

    if (wpNextWord)
    {
      *wpNextWord=wpCount;
    }
    if (bQuote) *bQuote=FALSE;
  }
  nWordLen=min(wpCount - wpText, nWordLenMax - 1);
  memcpy(wszWord, wpText, nWordLen * sizeof(wchar_t));
  wszWord[nWordLen]='\0';

  return nWordLen;
}

BOOL NextLineA(char **pText)
{
  while (**pText != '\r' && **pText != '\0') ++*pText;
  if (**pText == '\0') return FALSE;
  if (*++*pText == '\n') ++*pText;
  return TRUE;
}

BOOL NextLineW(wchar_t **wpText)
{
  while (**wpText != '\r' && **wpText != '\0') ++*wpText;
  if (**wpText == '\0') return FALSE;
  if (*++*wpText == '\n') ++*wpText;
  return TRUE;
}

BOOL SkipCommentA(char **pText)
{
  while (1)
  {
    while (**pText == ' ' || **pText == '\t' || **pText == '\r' || **pText == '\n') ++*pText;

    if (**pText == ';')
    {
      while (**pText != '\r' && **pText != '\0') ++*pText;
    }
    else break;
  }
  if (**pText == '\0')
    return FALSE;
  return TRUE;
}

BOOL SkipCommentW(wchar_t **wpText)
{
  while (1)
  {
    while (**wpText == ' ' || **wpText == '\t' || **wpText == '\r' || **wpText == '\n') ++*wpText;

    if (**wpText == ';')
    {
      while (**wpText != '\r' && **wpText != '\0') ++*wpText;
    }
    else break;
  }
  if (**wpText == '\0')
    return FALSE;
  return TRUE;
}

BOOL IsDelimiter(const wchar_t *s, wchar_t c)
{
  if (c == L'\n')
    return TRUE;

  while (*s && *s != c) ++s;

  if (*s)
    return TRUE;
  return FALSE;
}

char* GetFileExtA(char *pFile)
{
  int i;

  for (i=lstrlenA(pFile) - 1; i >= 0; --i)
  {
    if (pFile[i] == '.') return (pFile + i + 1);
    else if (pFile[i] == '\\') break;
  }
  return "";
}

wchar_t* GetFileExtW(wchar_t *wpFile)
{
  int i;

  for (i=lstrlenW(wpFile) - 1; i >= 0; --i)
  {
    if (wpFile[i] == '.') return (wpFile + i + 1);
    else if (wpFile[i] == '\\') break;
  }
  return L"";
}


//// Options

void ReadSyntaxFilesA()
{
  WIN32_FIND_DATAA wfd;
  HANDLE hSearch;
  HANDLE hFile;
  char szFindFiles[MAX_PATH];
  char szCodeFoldFile[MAX_PATH];

  //Start search for *.codefold
  wsprintfA(szFindFiles, "%s\\*.codefold", szCodeFoldDir);

  if ((hSearch=FindFirstFileA(szFindFiles, &wfd)) != INVALID_HANDLE_VALUE)
  {
    do
    {
      wsprintfA(szCodeFoldFile, "%s\\%s", szCodeFoldDir, wfd.cFileName);
      if ((hFile=CreateFileA(szCodeFoldFile, GENERIC_READ, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN, NULL)) != INVALID_HANDLE_VALUE)
      {
        StackLoadSyntaxFile(hFile, &hSyntaxFilesStack, NULL);
        CloseHandle(hFile);
      }
    }
    while (FindNextFileA(hSearch, &wfd));

    FindClose(hSearch);
  }
}

void ReadSyntaxFilesW()
{
  WIN32_FIND_DATAW wfd;
  HANDLE hSearch;
  HANDLE hFile;
  wchar_t wszFindFiles[MAX_PATH];
  wchar_t wszCodeFoldFile[MAX_PATH];

  //Start search for *.codefold
  wsprintfW(wszFindFiles, L"%s\\*.codefold", wszCodeFoldDir);

  if ((hSearch=FindFirstFileW(wszFindFiles, &wfd)) != INVALID_HANDLE_VALUE)
  {
    do
    {
      wsprintfW(wszCodeFoldFile, L"%s\\%s", wszCodeFoldDir, wfd.cFileName);

      if ((hFile=CreateFileW(wszCodeFoldFile, GENERIC_READ, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN, NULL)) != INVALID_HANDLE_VALUE)
      {
        StackLoadSyntaxFile(hFile, &hSyntaxFilesStack, NULL);
        CloseHandle(hFile);
      }
    }
    while (FindNextFileW(hSearch, &wfd));

    FindClose(hSearch);
  }
}

void ReadOptionsA()
{
  PLUGINOPTIONA po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_READ, (LPARAM)szPluginName))
  {
    po.pOptionName="DockSide";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nDockSide;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="DockRect";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&rcDockRect;
    po.dwData=sizeof(RECT);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="ShowDock";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bShowDock;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="FollowCaret";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bFollowCaret;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="NoPrintCollapsed";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bNoPrintCollapsed;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="PanelFirstBackground";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&crPanelFirstBackground;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="PanelSecondBackground";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&crPanelSecondBackground;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="PanelNormalFold";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&crPanelNormalFold;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="PanelActiveFold";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&crPanelActiveFold;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="CurrentCollapse";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwCurrentCollapse;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="CurrentGoBegin";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwCurrentGoBegin;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="CurrentGoEnd";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwCurrentGoEnd;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="CurrentSelect";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwCurrentSelect;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="AllCollapse";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwAllCollapse;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="AllNextLevel";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwAllNextLevel;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="AllPrevLevel";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwAllPrevLevel;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void ReadOptionsW()
{
  PLUGINOPTIONW po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_READ, (LPARAM)wszPluginName))
  {
    po.pOptionName=L"DockSide";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nDockSide;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"DockRect";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&rcDockRect;
    po.dwData=sizeof(RECT);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"ShowDock";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bShowDock;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"FollowCaret";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bFollowCaret;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"NoPrintCollapsed";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bNoPrintCollapsed;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"PanelFirstBackground";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&crPanelFirstBackground;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"PanelSecondBackground";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&crPanelSecondBackground;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"PanelNormalFold";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&crPanelNormalFold;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"PanelActiveFold";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&crPanelActiveFold;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"CurrentCollapse";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwCurrentCollapse;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"CurrentGoBegin";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwCurrentGoBegin;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"CurrentGoEnd";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwCurrentGoEnd;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"CurrentSelect";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwCurrentSelect;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"AllCollapse";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwAllCollapse;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"AllNextLevel";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwAllNextLevel;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"AllPrevLevel";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwAllPrevLevel;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void SaveOptionsA(DWORD dwFlags)
{
  PLUGINOPTIONA po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_SAVE, (LPARAM)szPluginName))
  {
    if (dwFlags & OF_WINDOWRECT)
    {
      po.pOptionName="DockSide";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&dkMainDlg->nSide;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="DockRect";
      po.dwType=PO_BINARY;
      po.lpData=(LPBYTE)&dkMainDlg->rcSize;
      po.dwData=sizeof(RECT);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
    }

    if (dwFlags & OF_SETTINGS)
    {
      po.pOptionName="ShowDock";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&bShowDock;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="FollowCaret";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&bFollowCaret;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="NoPrintCollapsed";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&bNoPrintCollapsed;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="PanelFirstBackground";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&crPanelFirstBackground;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="PanelSecondBackground";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&crPanelSecondBackground;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="PanelNormalFold";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&crPanelNormalFold;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="PanelActiveFold";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&crPanelActiveFold;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="CurrentCollapse";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&dwCurrentCollapse;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="CurrentGoBegin";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&dwCurrentGoBegin;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="CurrentGoEnd";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&dwCurrentGoEnd;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="CurrentSelect";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&dwCurrentSelect;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="AllCollapse";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&dwAllCollapse;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="AllNextLevel";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&dwAllNextLevel;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="AllPrevLevel";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&dwAllPrevLevel;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
    }

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void SaveOptionsW(DWORD dwFlags)
{
  PLUGINOPTIONW po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_SAVE, (LPARAM)wszPluginName))
  {
    if (dwFlags & OF_WINDOWRECT)
    {
      po.pOptionName=L"DockSide";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&dkMainDlg->nSide;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"DockRect";
      po.dwType=PO_BINARY;
      po.lpData=(LPBYTE)&dkMainDlg->rcSize;
      po.dwData=sizeof(RECT);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
    }

    if (dwFlags & OF_SETTINGS)
    {
      po.pOptionName=L"ShowDock";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&bShowDock;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"FollowCaret";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&bFollowCaret;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"NoPrintCollapsed";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&bNoPrintCollapsed;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"PanelFirstBackground";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&crPanelFirstBackground;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"PanelSecondBackground";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&crPanelSecondBackground;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"PanelNormalFold";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&crPanelNormalFold;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"PanelActiveFold";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&crPanelActiveFold;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"CurrentCollapse";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&dwCurrentCollapse;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"CurrentGoBegin";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&dwCurrentGoBegin;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"CurrentGoEnd";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&dwCurrentGoEnd;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"CurrentSelect";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&dwCurrentSelect;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"AllNextLevel";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&dwAllNextLevel;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"AllPrevLevel";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&dwAllPrevLevel;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
    }

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

char* GetLangStringA(LANGID wLangID, int nStringID)
{
  static char szStringBuf[MAX_PATH];

  WideCharToMultiByte(CP_ACP, 0, GetLangStringW(wLangID, nStringID), -1, szStringBuf, MAX_PATH, NULL, NULL);
  return szStringBuf;
}

wchar_t* GetLangStringW(LANGID wLangID, int nStringID)
{
  if (wLangID == LANG_RUSSIAN)
  {
    if (nStringID == STRID_GOBEGIN)
      return L"\x041F\x0435\x0440\x0435\x0439\x0442\x0438\x0020\x043A\x0020\x043D\x0430\x0447\x0430\x043B\x0443";
    if (nStringID == STRID_GOEND)
      return L"\x041F\x0435\x0440\x0435\x0439\x0442\x0438\x0020\x043A\x0020\x043A\x043E\x043D\x0446\x0443";
    if (nStringID == STRID_SELECT)
      return L"\x0412\x044B\x0434\x0435\x043B\x0438\x0442\x044C";
    if (nStringID == STRID_INVIEW)
      return L"\x041F\x043E\x043A\x0430\x0437\x0430\x0442\x044C";
    if (nStringID == STRID_LISTEXPAND)
      return L"\x0420\x0430\x0437\x0432\x0435\x0440\x043D\x0443\x0442\x044C\x0020\x0441\x043F\x0438\x0441\x043E\x043A";
    if (nStringID == STRID_LISTCOLLAPSE)
      return L"\x0421\x0432\x0435\x0440\x043D\x0443\x0442\x044C\x0020\x0441\x043F\x0438\x0441\x043E\x043A";
    if (nStringID == STRID_SETUP)
      return L"\x041D\x0430\x0441\x0442\x0440\x043E\x0439\x043A\x0438\x002E\x002E\x002E";
    if (nStringID == STRID_DOCKWINDOW)
      return L"\x041F\x043B\x0430\x0432\x0430\x044E\x0449\x0435\x0435\x0020\x043E\x043A\x043D\x043E";
    if (nStringID == STRID_SHOWDOCK)
      return L"\x041F\x043E\x043A\x0430\x0437\x044B\x0432\x0430\x0442\x044C\x0020\x0441\x043F\x0438\x0441\x043E\x043A";
    if (nStringID == STRID_FOLLOWCARET)
      return L"\x0421\x043B\x0435\x0434\x043E\x0432\x0430\x0442\x044C\x0020\x0437\x0430\x0020\x043A\x0430\x0440\x0435\x0442\x043A\x043E\x0439";
    if (nStringID == STRID_OTHERSETTINGS)
      return L"\x0414\x0440\x0443\x0433\x0438\x0435\x0020\x043D\x0430\x0441\x0442\x0440\x043E\x0439\x043A\x0438";
    if (nStringID == STRID_NOPRINTCOLLAPSED)
      return L"\x041D\x0435\x0020\x043F\x0435\x0447\x0430\x0442\x0430\x0442\x044C\x0020\x0441\x0432\x0435\x0440\x043D\x0443\x0442\x044B\x0435\x0020\x0441\x0442\x0440\x043E\x043A\x0438";
    if (nStringID == STRID_COLORS)
      return L"\x0426\x0432\x0435\x0442\x0430";
    if (nStringID == STRID_FIRSTBACKGROUNDCOLOR)
      return L"\x041F\x0435\x0440\x0432\x044B\x0439\x0020\x0444\x043E\x043D";
    if (nStringID == STRID_SECONDBACKGROUNDCOLOR)
      return L"\x0412\x0442\x043E\x0440\x043E\x0439\x0020\x0444\x043E\x043D";
    if (nStringID == STRID_NORMALFOLDCOLOR)
      return L"\x041E\x0431\x044B\x0447\x043D\x044B\x0439\x0020\x0443\x0440\x043E\x0432\x0435\x043D\x044C";
    if (nStringID == STRID_ACTIVEFOLDCOLOR)
      return L"\x0410\x043A\x0442\x0438\x0432\x043D\x044B\x0439\x0020\x0443\x0440\x043E\x0432\x0435\x043D\x044C";
    if (nStringID == STRID_HOTKEYCURRENT)
      return L"\x0422\x0435\x043A\x0443\x0449\x0438\x0439\x0020\x0443\x0440\x043E\x0432\x0435\x043D\x044C";
    if (nStringID == STRID_HOTKEYALL)
      return L"\x0412\x0441\x0435\x0020\x0443\x0440\x043E\x0432\x043D\x0438";
    if (nStringID == STRID_COLLAPSE)
      return L"\x0421\x0432\x0435\x0440\x043D\x0443\x0442\x044C";
    if (nStringID == STRID_NEXTLEVEL)
      return L"\x041A\x0020\x0441\x043B\x0435\x0434\x0443\x044E\x0449\x0435\x043C\x0443";
    if (nStringID == STRID_PREVLEVEL)
      return L"\x041A\x0020\x043F\x0440\x0435\x0434\x044B\x0434\x0443\x0449\x0435\x043C\x0443";
    if (nStringID == STRID_OK)
      return L"\x004F\x004B";
    if (nStringID == STRID_CANCEL)
      return L"\x041E\x0442\x043C\x0435\x043D\x0430";
  }
  else
  {
    if (nStringID == STRID_GOBEGIN)
      return L"Go to begin";
    if (nStringID == STRID_GOEND)
      return L"Go to end";
    if (nStringID == STRID_SELECT)
      return L"Select";
    if (nStringID == STRID_INVIEW)
      return L"In view";
    if (nStringID == STRID_LISTEXPAND)
      return L"Expand list";
    if (nStringID == STRID_LISTCOLLAPSE)
      return L"Collapse list";
    if (nStringID == STRID_SETUP)
      return L"Settings...";
    if (nStringID == STRID_DOCKWINDOW)
      return L"Dock window";
    if (nStringID == STRID_SHOWDOCK)
      return L"Show list";
    if (nStringID == STRID_FOLLOWCARET)
      return L"Follow caret";
    if (nStringID == STRID_OTHERSETTINGS)
      return L"Other settings";
    if (nStringID == STRID_NOPRINTCOLLAPSED)
      return L"No print collapsed lines";
    if (nStringID == STRID_COLORS)
      return L"Colors";
    if (nStringID == STRID_FIRSTBACKGROUNDCOLOR)
      return L"First background";
    if (nStringID == STRID_SECONDBACKGROUNDCOLOR)
      return L"Second background";
    if (nStringID == STRID_NORMALFOLDCOLOR)
      return L"Normal fold";
    if (nStringID == STRID_ACTIVEFOLDCOLOR)
      return L"Active fold";
    if (nStringID == STRID_HOTKEYCURRENT)
      return L"Current level";
    if (nStringID == STRID_HOTKEYALL)
      return L"All levels";
    if (nStringID == STRID_COLLAPSE)
      return L"Collapse";
    if (nStringID == STRID_NEXTLEVEL)
      return L"Next level";
    if (nStringID == STRID_PREVLEVEL)
      return L"Previous level";
    if (nStringID == STRID_OK)
      return L"OK";
    if (nStringID == STRID_CANCEL)
      return L"Cancel";
  }
  return L"";
}

BOOL IsExtCallParamValid(LPARAM lParam, int nIndex)
{
  if (*((int *)lParam) >= (int)((nIndex + 1) * sizeof(int)))
    return TRUE;
  return FALSE;
}

int GetExtCallParam(LPARAM lParam, int nIndex)
{
  if (*((int *)lParam) >= (int)((nIndex + 1) * sizeof(int)))
    return *(((int *)lParam) + nIndex);
  return 0;
}

void InitCommon(PLUGINDATA *pd)
{
  bInitCommon=TRUE;
  hInstanceDLL=pd->hInstanceDLL;
  hMainWnd=pd->hMainWnd;
  hWndEdit=pd->hWndEdit;
  hMdiClient=pd->hMdiClient;
  hMainIcon=pd->hMainIcon;
  bOldWindows=pd->bOldWindows;
  bMDI=pd->bMDI;
  wLangSystem=PRIMARYLANGID(pd->wLangSystem);

  if (bOldWindows)
  {
    int i;

    for (i=0; ((char *)pd->pFunction)[i] != ':'; ++i)
      szPluginName[i]=((char *)pd->pFunction)[i];
    szPluginName[i]='\0';
  }
  else
  {
    int i;

    for (i=0; ((wchar_t *)pd->pFunction)[i] != ':'; ++i)
      wszPluginName[i]=((wchar_t *)pd->pFunction)[i];
    wszPluginName[i]='\0';
  }

  if (bOldWindows)
  {
    wsprintfA(szCodeFoldDir, "%s\\AkelFiles\\Plugs\\CodeFold", (char *)pd->pAkelDir);
    ReadOptionsA();
  }
  else
  {
    wsprintfW(wszCodeFoldDir, L"%s\\AkelFiles\\Plugs\\CodeFold", (wchar_t *)pd->pAkelDir);
    ReadOptionsW();
  }
}

void InitMain()
{
  bInitMain=TRUE;

  if (bOldWindows)
    ReadSyntaxFilesA();
  else
    ReadSyntaxFilesW();

  //Add hotkeys
  if (bOldWindows)
  {
    PLUGINFUNCTIONA pf={0};

    if (dwCurrentCollapse && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwCurrentCollapse))
    {
      pf.nFunctionLen=wsprintfA(pf.szFunction, "CodeFold::Main::CurrentCollapse");
      pf.wHotkey=(WORD)dwCurrentCollapse;
      pf.PluginProc=CurrentCollapseProc;
      pfaCurrentCollapse=(PLUGINFUNCTIONA *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
    if (dwCurrentGoBegin && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwCurrentGoBegin))
    {
      pf.nFunctionLen=wsprintfA(pf.szFunction, "CodeFold::Main::CurrentGoBegin");
      pf.wHotkey=(WORD)dwCurrentGoBegin;
      pf.PluginProc=CurrentGoBeginProc;
      pfaCurrentGoBegin=(PLUGINFUNCTIONA *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
    if (dwCurrentGoEnd && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwCurrentGoEnd))
    {
      pf.nFunctionLen=wsprintfA(pf.szFunction, "CodeFold::Main::CurrentGoEnd");
      pf.wHotkey=(WORD)dwCurrentGoEnd;
      pf.PluginProc=CurrentGoEndProc;
      pfaCurrentGoEnd=(PLUGINFUNCTIONA *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
    if (dwCurrentSelect && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwCurrentSelect))
    {
      pf.nFunctionLen=wsprintfA(pf.szFunction, "CodeFold::Main::CurrentSelect");
      pf.wHotkey=(WORD)dwCurrentSelect;
      pf.PluginProc=CurrentSelectProc;
      pfaCurrentSelect=(PLUGINFUNCTIONA *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
    if (dwAllCollapse && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwAllCollapse))
    {
      pf.nFunctionLen=wsprintfA(pf.szFunction, "CodeFold::Main::AllCollapse");
      pf.wHotkey=(WORD)dwAllCollapse;
      pf.PluginProc=AllCollapseProc;
      pfaAllCollapse=(PLUGINFUNCTIONA *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
    if (dwAllNextLevel && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwAllNextLevel))
    {
      pf.nFunctionLen=wsprintfA(pf.szFunction, "CodeFold::Main::AllNextLevel");
      pf.wHotkey=(WORD)dwAllNextLevel;
      pf.PluginProc=AllNextLevelProc;
      pfaAllNextLevel=(PLUGINFUNCTIONA *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
    if (dwAllPrevLevel && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwAllPrevLevel))
    {
      pf.nFunctionLen=wsprintfA(pf.szFunction, "CodeFold::Main::AllPrevLevel");
      pf.wHotkey=(WORD)dwAllPrevLevel;
      pf.PluginProc=AllPrevLevelProc;
      pfaAllPrevLevel=(PLUGINFUNCTIONA *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
  }
  else
  {
    PLUGINFUNCTIONW pf={0};

    if (dwCurrentCollapse && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwCurrentCollapse))
    {
      pf.nFunctionLen=wsprintfW(pf.szFunction, L"CodeFold::Main::CurrentCollapse");
      pf.wHotkey=(WORD)dwCurrentCollapse;
      pf.PluginProc=CurrentCollapseProc;
      pfwCurrentCollapse=(PLUGINFUNCTIONW *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
    if (dwCurrentGoBegin && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwCurrentGoBegin))
    {
      pf.nFunctionLen=wsprintfW(pf.szFunction, L"CodeFold::Main::CurrentGoBegin");
      pf.wHotkey=(WORD)dwCurrentGoBegin;
      pf.PluginProc=CurrentGoBeginProc;
      pfwCurrentGoBegin=(PLUGINFUNCTIONW *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
    if (dwCurrentGoEnd && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwCurrentGoEnd))
    {
      pf.nFunctionLen=wsprintfW(pf.szFunction, L"CodeFold::Main::CurrentGoEnd");
      pf.wHotkey=(WORD)dwCurrentGoEnd;
      pf.PluginProc=CurrentGoEndProc;
      pfwCurrentGoEnd=(PLUGINFUNCTIONW *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
    if (dwCurrentSelect && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwCurrentSelect))
    {
      pf.nFunctionLen=wsprintfW(pf.szFunction, L"CodeFold::Main::CurrentSelect");
      pf.wHotkey=(WORD)dwCurrentSelect;
      pf.PluginProc=CurrentSelectProc;
      pfwCurrentSelect=(PLUGINFUNCTIONW *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
    if (dwAllCollapse && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwAllCollapse))
    {
      pf.nFunctionLen=wsprintfW(pf.szFunction, L"CodeFold::Main::AllCollapse");
      pf.wHotkey=(WORD)dwAllCollapse;
      pf.PluginProc=AllCollapseProc;
      pfwAllCollapse=(PLUGINFUNCTIONW *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
    if (dwAllNextLevel && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwAllNextLevel))
    {
      pf.nFunctionLen=wsprintfW(pf.szFunction, L"CodeFold::Main::AllNextLevel");
      pf.wHotkey=(WORD)dwAllNextLevel;
      pf.PluginProc=AllNextLevelProc;
      pfwAllNextLevel=(PLUGINFUNCTIONW *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
    if (dwAllPrevLevel && !SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, dwAllPrevLevel))
    {
      pf.nFunctionLen=wsprintfW(pf.szFunction, L"CodeFold::Main::AllPrevLevel");
      pf.wHotkey=(WORD)dwAllPrevLevel;
      pf.PluginProc=AllPrevLevelProc;
      pfwAllPrevLevel=(PLUGINFUNCTIONW *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf);
    }
  }

  //SubClass
  NewMainProcData=NULL;
  SendMessage(hMainWnd, AKD_SETMAINPROC, (WPARAM)NewMainProc, (LPARAM)&NewMainProcData);

  if (bMDI)
  {
    NewFrameProcData=NULL;
    SendMessage(hMainWnd, AKD_SETFRAMEPROC, (WPARAM)NewFrameProc, (LPARAM)&NewFrameProcData);
  }

  NewEditProcData=NULL;
  SendMessage(hMainWnd, AKD_SETEDITPROC, (WPARAM)NewEditProc, (LPARAM)&NewEditProcData);
}

void UninitMain()
{
  bInitMain=FALSE;

  //Remove hotkeys
  if (bOldWindows)
  {
    if (pfaCurrentCollapse)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfaCurrentCollapse);
      pfaCurrentCollapse=NULL;
    }
    if (pfaCurrentGoBegin)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfaCurrentGoBegin);
      pfaCurrentGoBegin=NULL;
    }
    if (pfaCurrentGoEnd)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfaCurrentGoEnd);
      pfaCurrentGoEnd=NULL;
    }
    if (pfaCurrentSelect)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfaCurrentSelect);
      pfaCurrentSelect=NULL;
    }
    if (pfaAllCollapse)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfaAllCollapse);
      pfaAllCollapse=NULL;
    }
    if (pfaAllNextLevel)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfaAllNextLevel);
      pfaAllNextLevel=NULL;
    }
    if (pfaAllPrevLevel)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfaAllPrevLevel);
      pfaAllPrevLevel=NULL;
    }
  }
  else
  {
    if (pfwCurrentCollapse)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfwCurrentCollapse);
      pfwCurrentCollapse=NULL;
    }
    if (pfwCurrentGoBegin)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfwCurrentGoBegin);
      pfwCurrentGoBegin=NULL;
    }
    if (pfwCurrentGoEnd)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfwCurrentGoEnd);
      pfwCurrentGoEnd=NULL;
    }
    if (pfwCurrentSelect)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfwCurrentSelect);
      pfwCurrentSelect=NULL;
    }
    if (pfwAllCollapse)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfwAllCollapse);
      pfwAllCollapse=NULL;
    }
    if (pfwAllNextLevel)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfwAllNextLevel);
      pfwAllNextLevel=NULL;
    }
    if (pfwAllPrevLevel)
    {
      SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfwAllPrevLevel);
      pfwAllPrevLevel=NULL;
    }
  }

  //Remove subclass
  if (NewMainProcData)
  {
    SendMessage(hMainWnd, AKD_SETMAINPROC, (WPARAM)NULL, (LPARAM)&NewMainProcData);
    NewMainProcData=NULL;
  }
  if (NewFrameProcData)
  {
    SendMessage(hMainWnd, AKD_SETFRAMEPROC, (WPARAM)NULL, (LPARAM)&NewFrameProcData);
    NewFrameProcData=NULL;
  }
  if (NewEditProcData)
  {
    SendMessage(hMainWnd, AKD_SETEDITPROC, (WPARAM)NULL, (LPARAM)&NewEditProcData);
    NewEditProcData=NULL;
  }
  StackFreeSyntaxFile(&hSyntaxFilesStack);
}

//Entry point
extern "C" BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
  if (fdwReason == DLL_PROCESS_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_DETACH)
  {
  }
  else if (fdwReason == DLL_PROCESS_DETACH)
  {
  }
  return TRUE;
}
